/*
 * $RCSfile: crypt.h,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 *
 *
 * $Log: crypt.h,v $
 * Revision 1.9  1995/12/01 05:10:23  spatsch
 * Added rfc1829
 *
 * Revision 1.8  1995/10/28  01:20:12  rcs
 * Added IV_ON_UPMSG flag for Photuris.
 *
 * Revision 1.7  1995/10/19  02:08:16  rcs
 * Added flag to take IV from first 8 bytes of outgoing message.
 *
 * Revision 1.5  1995/05/04  23:03:23  ho
 * Added IPRIVATE query for rlogin queries.
 * .
 *
 * Revision 1.4  1994/04/29  00:10:44  rcs
 * Added new control op CRYPT_IVEC_FROM_KEY.
 *
 * Revision 1.3  1994/02/05  00:06:13  menze
 *   [ 1994/01/28          menze ]
 *   Added prototype for tcrypt_init
 *
 */

#ifndef crypt_h
#define crypt_h



#define CRYPT_IN_RESOLVE      (CRYPT_CTL*MAXOPS + 0)
#define CRYPT_IN_SET          (CRYPT_CTL*MAXOPS + 1)
#define CRYPT_IN_INVALIDATE   (CRYPT_CTL*MAXOPS + 2)
#define CRYPT_IN_ISEMPTY      (CRYPT_CTL*MAXOPS + 3)
#define CRYPT_IN_ISVALID      (CRYPT_CTL*MAXOPS + 4)
#define CRYPT_IN_ISINVALID    (CRYPT_CTL*MAXOPS + 5)
#define CRYPT_IN_KEYSIZE      (CRYPT_CTL*MAXOPS + 6)
#define CRYPT_IN_HOSTSIZE     (CRYPT_CTL*MAXOPS + 7)
#define CRYPT_OUT_RESOLVE     (CRYPT_CTL*MAXOPS + 8)
#define CRYPT_OUT_SET         (CRYPT_CTL*MAXOPS + 9)
#define CRYPT_OUT_INVALIDATE  (CRYPT_CTL*MAXOPS + 10)
#define CRYPT_OUT_ISEMPTY     (CRYPT_CTL*MAXOPS + 11)
#define CRYPT_OUT_ISVALID     (CRYPT_CTL*MAXOPS + 12)
#define CRYPT_OUT_ISINVALID   (CRYPT_CTL*MAXOPS + 13)
#define CRYPT_OUT_KEYSIZE     (CRYPT_CTL*MAXOPS + 14)
#define CRYPT_OUT_HOSTSIZE    (CRYPT_CTL*MAXOPS + 15)
#define CRYPT_IVEC_FROM_KEY   (CRYPT_CTL*MAXOPS + 16)
#define CRYPT_ISPRIVATE       (CRYPT_CTL*MAXOPS + 17)
#define CRYPT_READ_FLAGS      (CRYPT_CTL*MAXOPS + 18)
#define CRYPT_SET_FLAGS       (CRYPT_CTL*MAXOPS + 19)
#define CRYPT_CLEAR_FLAGS     (CRYPT_CTL*MAXOPS + 20)
#define CRYPT_READ_IV_DOWN    (CRYPT_CTL*MAXOPS + 21)
#define CRYPT_SET_IV_DOWN     (CRYPT_CTL*MAXOPS + 22)
#define CRYPT_READ_IV_UP      (CRYPT_CTL*MAXOPS + 23)
#define CRYPT_SET_IV_UP       (CRYPT_CTL*MAXOPS + 24)
#define CRYPT_READ_IV_RANDOM  (CRYPT_CTL*MAXOPS + 25)
#define CRYPT_SET_IV_RANDOM   (CRYPT_CTL*MAXOPS + 26)
#define CRYPT_READ_FRAG_DOWN  (CRYPT_CTL*MAXOPS + 27)
#define CRYPT_SET_FRAG_DOWN   (CRYPT_CTL*MAXOPS + 28)
#define CRYPT_READ_FRAG_UP    (CRYPT_CTL*MAXOPS + 29)
#define CRYPT_SET_FRAG_UP     (CRYPT_CTL*MAXOPS + 30)

/* Flags: public names */
#define CRYPT_IP_PSEUDOHDR_HACK  1  /* either protocol or session */
#define CRYPT_IV_FROM_KEY        2  /* protocol or session */
#define CRYPT_ALL_PARTS          4  /* protocol */
#define CRYPT_CBC                8  /* protocol */
#define CRYPT_DECRYPT           16  /* protocol */
#define CRYPT_USE_IPSEC         32  /* protocol */
#define CRYPT_SEND_IV           64  /* protocol or session */
#define CRYPT_SEND_IV_ONCE     128  /* protocol; copied to session */
#define CRYPT_RCV_IV_ONCE      256  /* protocol; copied to session */
#define CRYPT_CRYPT_IV         512  /* protocol or session */
#define CRYPT_ROTATE_BYTE     1024  /* protocol or session */
#define CRYPT_IV_STATE        2048  /* Save IV after encryption/decryption. PorS*/
#define CRYPT_RANDOM_IV       4096  /* protocol or session */
#define CRYPT_IV_FROM_MSG     8192  /* protocol or session */
#define CRYPT_IV_ON_UPMSG    16384  /* protocol or session */
#define CRYPT_FRAGMENTS_DOWN 32768  /* protocol or session */
#define CRYPT_FRAGMENTS_UP   65536  /* protocol or session */
#define CRYPT_NO_PADDING_CHECK 131072  /* protocol */
#define CRYPT_RFC_1829      262144  /* protocol */




#  ifdef __STDC__

void	crypt_init( XObj );
void	tcrypt_init( XObj );

#  endif


#endif
