/*
 * $RCSfile: bounce.c,v $

 * The BOUNCE protocol pushes all demuxed message back to the llp that
 * popped them.  It behaves the same way a wall does when you throw a
 * ball directly at it; hence the name.

 * The exceptions are xPush and xControl operations which are sent to
 * the first llp.  Imagine a wall that can also throw balls at you.

 * Copyright (c) 1996	Arizona Board of Regents
 *
 * $Log: bounce.c,v $
 * Revision 1.2  1996/07/23 22:13:40  mjk
 * *** empty log message ***
 *
 * Revision 1.1  1996/05/24  22:39:22  mjk
 * Initial revision
 *
 */


#include "xkernel.h"


void			bounce_init(XObj);
static xmsg_handle_t	bounce_push(XObj, Msg *);
static int		bounce_control(XObj, int, char *, int);
static xkern_return_t	bounce_demux(XObj, XObj, Msg *);
static xkern_return_t	bounce_openenable(XObj, XObj, XObj, Part *);

int	tracebouncep;


/* 
 *  Public Functions
 */

void
bounce_init(XObj self) 
{
  XObj	llp;
  Part	part;

  xTraceP0(self, TR_GROSS_EVENTS, "init");

  if ( !xIsProtocol(llp = xGetDown(self, 0)) ) {
    xError("bounce can not get lower protocol");
    return;
  }

  partInit(&part, 1);
  partPush(part, ANY_HOST, 0);
  if ( xOpenEnable(self, self, llp, &part) == XK_FAILURE ) {
    xError("bounce cannot openenable lower protocol");
    return;
  }
  
  self->openenable  = bounce_openenable;
  self->control	    = bounce_control;
  self->push	    = bounce_push;
  self->demux       = bounce_demux;
} /* bounce_init */



/* 
 * Public Protocol Methods 
 */

static xkern_return_t
bounce_openenable(XObj self, XObj hlpRcv, XObj hlpType, Part *p)
{
  return XK_SUCCESS;
} /* bounce_openenable */


static int
bounce_control(XObj self, int opcode, char *buf, int buflen)
{
  return xControl(xGetDown(self, 0), opcode, buf, buflen);
} /* bounce_control */


static xmsg_handle_t
bounce_push(XObj self, Msg *packet)
{
  return xPush(xGetDown(self, 0), packet);
} /* bounce_push */


static xkern_return_t
bounce_demux(XObj self, XObj lls, Msg *packet)
{
  return xPush(lls, packet);
} /* bounce_demux */


/* bounce.c ends here */
 


