/*
 * $RCSfile: flood.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 *
 * $Log: flood.c,v $
 * Revision 1.4  1996/05/28 18:13:05  ho
 * Removed bogus token concatenation for include file pathname.
 * Changed Makefile (why doesn't this work "naturally"?)
 *
 * Revision 1.3  1996/05/26  22:40:44  ho
 * pointer casts for bcopy.
 *
 * Revision 1.2  1996/05/01  21:33:13  rajesh
 * changes for flood to use multicast addresses.
 *
 *
 */

#include "xkernel.h"
#include "ip.h"                /* For IP pseudoheader fixup. */
#include "flood_i.h"
#include "floodhdr.h"

int tracefloodp;

static XObj		floodCreateSessn( XObj, XObj, XObj, XObj);
static void		floodSessnInit( XObj );
static int		floodControlProtl( XObj, int, char *, int );
static xkern_return_t	floodDemux( XObj, XObj, Msg * );
static XObj		floodOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	floodOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	floodOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	floodOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t	floodClose( XObj );
static xmsg_handle_t	floodPush( XObj, Msg * );
static xkern_return_t	floodPop( XObj, XObj, Msg *, void * );
static int		floodControlSessn( XObj, int, char *, int );

extern void             flood_init( XObj );
static xkern_return_t   romaddr( XObj, char**, int,int, void* );

void build_participants( IPhost* interface, Part* plist );
void do_flood( XObj, Msg*, Part*, IPhost* );

long flood_hdr_load(  VOID* hdr, char* src, long len, VOID* arg );
void flood_hdr_store( VOID* hdr, char* dst, long len, VOID* arg );

void flood_log_init( XObj protl );
void flood_log_add( PState* ps, flood_stamp_t* stamp );
int  flood_log_find( PState* ps, flood_stamp_t* stamp );
void flood_log_update( Event ev, void* arg );

/* must_flood( ps, addr )
 * 
 * This is just a shorthand for the test that has to be done to verify if
 * the address presented (usually the destination address in an Open or
 * OpenDone) is in the list of flooding addresses.
 */


#define must_flood( ps, addr )   \
  ( mapResolve(((PState*)(ps))->flooding_addresses,addr, NULL ) == XK_SUCCESS )

/*
 * build_participants(interface,plist)
 *
 * Given the address of an interface through which a message must be sent,
 * returns (in plist) a participant list that can be used to open the
 * session needed to send it.
 *
 * This right now uses the broadcast address for each interface. When
 * support for class D addresses is working, it may just replace the
 * local participant by the new interface.
 */

void
build_participants( IPhost* interface, Part* plist )
{
  static IPhost bcast;
  IPhost        mask, cmask;

  /* Plist is *a copy* of the participant list used to open the session,
   * so first it must pop the current top, that may need to be changed.
   */

  partPop(plist[1]);                                 /* old interface addr */
  partPop(plist[0]);                                 /* bcast addr for it  */

  /* The following probably will not need to be changed to work with class D
   * addresses. The local participant in any case must be the address of the
   * interface that will be used.
   */

  partPush(plist[1],(void*)interface,sizeof(*interface));   /* New interface */

  /* The remote participant right now is just the broadcast address associated
   * with the interface given. When using class D addresses, this is just the
   * same address as the one used in the initial session.
   */

  netMaskFind( &mask, interface );
  IP_COMP( cmask, mask );
  IP_OR( bcast, cmask, *interface );                        /* Compute bcast */

  partPush(plist[0],(void*)&bcast,sizeof(bcast));           /* complete list */
  xTrace3(floodp,TR_MAJOR_EVENTS,"flood: build_participants for %s: (%s,%s)",
	 ipHostStr(interface),ipHostStr(interface),ipHostStr(&bcast));
}

/*
 * flood( iflist, ifcnt, msg, plist, interface )
 *
 * Given the list of interfaces, a message and an original participant
 * list, send the message to all interfaces in the list, except to
 * "interface". This is usually the interface through which the message
 * was received, or the interface used in the initial open.
 */

void
do_flood( XObj sessn, Msg* msg, Part* plist, IPhost* interface )
{
  XObj protl = xMyProtl(sessn);
  PState* ps = (PState*)protl->state;
  int i;
  IPhost* destination;
  IPhost* source;
  Msg duplicate;
  Part p[2];
  XObj lls;

  xTrace0(floodp,TR_MAJOR_EVENTS,"do_flood: entering");

  /* In order to open sessions for all interfaces, the participant list has
   * to be reused and altered for each interface. To do that, we will make
   * a copy of the original one to restore it before it is used for each 
   * interface. After the copy is done, the original one is checked for
   * the interface and destination addresses.
   */

  /* NOTE NOTE NOTE NOTE NOTE NOTE:
   * ==============================
   * I am not sure how to deal with a variable participant list right now.
   * At this point, I assume all sessions have a remote and local participants,
   * what seems reasonable for the applications so far. Changes may be needed
   * to allow a third participant to specify security options, which must
   * be properly mirrored for all sessions.
   */
  bcopy((char *)plist,(char *)p,sizeof(p));   /* Freeze particpant list */
  destination = partPop(plist[0]);
  source      = partPop(plist[1]);

  /* Now we have saved the participant list, and have the message to be sent
   * and the list of all interfaces for this host, so it's just a matter of
   * iterating through them...
   */
  for (i=0;i<ps->if_count;++i) {

    /* First we may have to exclude some interfaces. First of all, if this
     * message is being "re-flooded" after being received at one interface,
     * we must at least exclude that interface (provided as a parameter).
     * 
     * Second, when the keymanager with information from the routing protocol
     * is added, a test must be done for each interface to see if it is
     * stored in the key manager. (Refer to the file flood.txt for a 
     * discussion of this.
     */
    if ( !interface || !( IP_EQUAL(ps->if_list[i],*interface) ) ){

      /* First of all, create a copy of the original message and restore
       * the original participant list. (Is the copy of the message needed?)
       */
      msgConstructCopy(&duplicate,msg);
      bcopy((char *)p,(char *)plist,sizeof(p));

      /* Once the original participants are restored, they must be altered
       * to hold the information to identify the new interface, in stead of
       * the original one.
       */
      build_participants( &(ps->if_list[i]), plist );
      xTrace1(floodp,TR_MAJOR_EVENTS,"do_flood: sending to %s",
	      ipHostStr(&(ps->if_list[i])));

      /* Now is just a matter of opening the session down with the new
       * participant list, pushing the message copy, closing the session
       * and destroying the message copy.
       */
      lls = xOpen(protl, xHlpType(sessn), xGetDown(protl, 0), plist);
      if ( !xIsSession(lls) ) {
	xTrace1(floodp,TR_ERRORS,"do_flood: unable to open session for"
		       " interface %s",ipHostStr(&(ps->if_list[i])) );
      }
      xPush(lls,&duplicate);
      xClose(lls);
      msgDestroy(&duplicate);
    }
  }

  bcopy((char *)p,(char *)plist,sizeof(p)); /* Restores participant list */
  xTrace0(floodp,TR_MAJOR_EVENTS,"do_flood: returning");
}

static XObjRomOpt rmopts[1] = { { "address", 1, romaddr } };

/*
 * romaddr - gets one address to be watched for flooding from ROM file
 *
 * Given the address, if no session exists for it a new one is created
 * (what automatically adds it to the active sessions map) and an
 * open enable is executed for that local address.
 */

static xkern_return_t
romaddr( XObj self, char **ptr, int  fields, int line, void *arg )
{
  PState* ps = (PState*) self->state;
  IPhost  addr;

  /* Now we just have to read the address from the rom file and add it
   * to the "flooding_addresses" mapping in the protocol state.
   */

  xTrace1(floodp, TR_MAJOR_EVENTS, "flood ADDRESS rom opt at line %d",line);

  if ( str2ipHost( &addr, ptr[2] ) == XK_FAILURE ) {
    xTrace2(floodp, TR_ERRORS, "invalid addr at line %d: %s", line, ptr[2] );
    return XK_FAILURE;
  }
  mapBind(ps->flooding_addresses,(void*)&addr,NULL);
  return XK_SUCCESS;
}

/*
 * flood_init:
 *
 * This is just general initialization, there are no great subtleties.
 */


void
flood_init( self )
     XObj self;
{
    PState	*ps;
    XObj	llp;
    Event       ev;

    xTrace0(floodp, TR_EVENTS, "flood_init entered");
    xAssert(xIsProtocol(self));
    
    if ( ! xIsProtocol(llp = xGetDown(self, 0)) ) {
      xError("flood_init: flood cannot get lower protocol object");
      return;
    }

    /* Set up pointers to the appropriate functions */
    self->control = floodControlProtl;
    self->open = floodOpen;
    self->openenable = floodOpenEnable;
    self->opendisable = floodOpenDisable;
    self->opendone = floodOpenDone;
    self->close = floodClose;  
    self->demux = floodDemux;
    self->push = floodPush;

    /* Allocate the protocol stae and initialize it: */
    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    ps->actMap = mapCreate(FLOOD_ACTIVE_MAP_SZ, sizeof(ActiveId));
    ps->pasMap = mapCreate(FLOOD_PASSIVE_MAP_SZ, sizeof(PassiveId));
    ps->flags = 0;

    /* Use the GETMYHOSTCOUNT and GETMYHOST operations to read the list
     * of interfaces in this host
     */
    xControl(llp,GETMYHOSTCOUNT,(void*)&(ps->if_count),sizeof(ps->if_count));
    if ( ( ps->if_count == 0 )||( ps->if_count > MAX_IF_COUNT ) ) 
    {
      xTrace1(floodp,TR_ERRORS,"flood_init: number of interfaces (%d)",
	      ps->if_count);
      xError("flood_init: unable to create protocol");
      return;
    }
    xControl(llp,GETMYHOST,(void*)(ps->if_list),sizeof(ps->if_list));

    /* Use the first address of the list as the host id in the stamp.
     * This should work, as long as addresses are not dynamic...
     */
    bcopy((char *)&(ps->if_list[0]),(char *)&(ps->next_stamp.host_addr),sizeof(IPhost));

    ps->next_stamp.seqn = 0;
    xTrace2(floodp, TR_MAJOR_EVENTS, "flood_init: stamp (%s,%d)",
	    ipHostStr(&(ps->next_stamp.host_addr)),ps->next_stamp.seqn);

	    
    /* The time wheel for logging recent stamps is initialized and a
     * periodic event is scheduled to perform the updating of the wheel and
     * discarding of old information.
     */
    flood_log_init( self );
    PS_CURR_SLOT(ps) = 0;
    ev = evSchedule( flood_log_update, (void*)self, FLOOD_LOG_INTERVAL*1000000);
    evDetach( ev );


    /* Create the mapping for addresses, that will be filled as Rom options */
    ps->flooding_addresses = mapCreate(FLOOD_ADDRS_MAP_SZ, sizeof(IPhost));

    findXObjRomOpts(self, rmopts, sizeof(rmopts)/sizeof(XObjRomOpt), 0);

    xTrace0(floodp, TR_EVENTS, "flood_init exiting");
}

/*
 * floodOpen: uses the participant list intended for IP to determine
 * if it is a flooding session or not, and set the appropriate flags.
 */

static XObj
floodOpen( XObj self, XObj hlpRcv, XObj hlpType, Part* part )
{
    XObj 	floodSessn;
    XObj	lls;
    PState	*pstate;
    int         flood_flag = 0;

    pstate = (PState *) self->state;
    
    xTrace4(floodp,TR_EVENTS,"floodOpen calling xOpen with args %x %x %x %x",
	    self,hlpType,xGetDown(self,0),part);

    /* First check the remote participant (needs to pop and push back) to
     * see if it must flood the session. Calls the function must_flood
     * for that.
     */
    {
      IPhost* host;
      int     len;
      len  = partStackTopByteLen( part[0] );
      host = (IPhost*) partPop( part[0] );
      partPush( part[0], host, len );
      if ( len != 0 ) {
        /* This test must be done here, before the participant list is
	 * passed down on xOpen, because that will alter it, making it
	 * unusable. We need to use a temporary variable (flood_flag)
	 * to store the result of must_flood until we can save it to
	 * the created session state.
	 */
	flood_flag = must_flood( pstate, host );
	xTrace2(floodp,TR_FULL_TRACE,"Open: %s(%d)",ipHostStr(host),flood_flag);
      }
    }

    /*
     * Now open lower session on IP. Then check if the returned session
     * already exists in the active map, in which case the newly created
     * reference may be closed. Otherwise, a new session is created and
     * added to the binding by floodCreateSessn.
     */

    lls = xOpen(self, hlpType, xGetDown(self, 0), part);
    if ( lls == ERR_XOBJ ) {
      xTrace0(floodp, TR_ERRORS, "floodOpen: could not open lls");
      return ERR_XOBJ;
    }

    if ( mapResolve(pstate->actMap, &lls, &floodSessn) == XK_SUCCESS ) {
        xTrace0(floodp, TR_FULL_TRACE, "floodOpen: found an existing one");
        xClose(lls);
    } else {
        xTrace0(floodp, TR_FULL_TRACE, "floodOpen: creating a new one");
        if ((floodSessn=floodCreateSessn(self,hlpRcv,hlpType,lls))==ERR_XOBJ) {
	  xTrace0(floodp, TR_ERRORS, "floodOpen: create session failed");
	  xClose(lls);
        }
	/* Only at this point we have the definitive session, and can set
	 * the state flags properly with the value previously computed.
	 */
	if ( flood_flag ) {
	  ((SState*) floodSessn->state)->flags |= 2;
	  xTrace0(floodp,TR_FULL_TRACE,"floodOpen: session marked for flooding");
	}
    }
    return (floodSessn);
}

/*
 * floodCreateSessn: Creates a new session based on the lower and upper
 * protocols/sessions. Flood behaves a little like a virtual protocol,
 * in which the sessions hlpRcv points to the upper protocol, what is
 * used to allow IP to demux messages based on the upper protocol's number.
 */
static XObj
floodCreateSessn( self, hlpRcv, hlpType, lls)
    XObj	self, hlpRcv, hlpType, lls;
{
    XObj	s;
    SState	*ss;
    PState	*ps = (PState *)self->state;
    XObj        SessnArray[1];

    xTrace4(floodp, TR_EVENTS, "floodCreateSessn entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, lls);

    /* Fist create session and try to add it to the active sessions map
     */
    SessnArray[0]=lls;
    s = xCreateSessn(floodSessnInit, hlpRcv, hlpType, self, 1, SessnArray);
    s->binding = (Bind) mapBind(ps->actMap, (char *)&lls, s);

    if ( s->binding == ERR_BIND ) {
	xTrace0(floodp, TR_ERRORS, "floodCreateSessn: mapBind error");
	xDestroy(s);
	return ERR_XOBJ;
      }

    /* Now create the session state and initialize it. Flags will be set
     * elsewhere.
     */
    ss = X_NEW(SState);
    xTrace1(floodp, TR_FULL_TRACE, "floodCreateSessn: X_NEW(SState) returned %x", ss);
    s->state = (VOID *)ss;
    ss->flags = 0;

    xTrace1(floodp, TR_MORE_EVENTS, "floodCreateSessn returning %x",s);
    return s;
}

/* 
 * floodOpenEnable: uses the defaultOpenEnable to make things simpler.
 * The key used for it is not the protocol number, as usual, but the
 * pointer to the upper protocol issuing the OpenEnable. This is
 * inherited from confounder.
 */

static xkern_return_t
floodOpenEnable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(floodp, TR_EVENTS, "floodOpenEnable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    /* As mentioned, it uses &hlpType as the key to defaultOpenEnable:
     */
    if ( defaultOpenEnable(ps->pasMap, hlpRcv, hlpType, 
	(VOID *)&hlpType) == XK_FAILURE ) {
        xTrace0(floodp, TR_MORE_EVENTS,
	               "floodOpenEnable defaultOpenEnable failed ");
        return XK_FAILURE;
    }

    /* Once that works, it may now call xOpenEnable on IP passing itself
     * and the higher level protocol, as a virtual protocol would do.
     */
    val = xOpenEnable(self, hlpType, xGetDown(self,0), part);
    if ( val == XK_FAILURE ) {
        xTrace0(floodp, TR_MORE_EVENTS, "floodOpenEnable xOpenEnable failed ");
        return XK_FAILURE;
    }
    
    xTrace1(floodp, TR_MORE_EVENTS, "floodOpenEnable returning %x", val);
    return val;
} 

/*
 * floodOpenDisable: Just the default code, with the detail of using the
 * higher protocol as the key to defaultOpenDisable.
 */
static xkern_return_t
floodOpenDisable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(floodp, TR_EVENTS, "floodOpenDisable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    val = defaultOpenDisable(ps->pasMap, hlpRcv, hlpType, (void*)&hlpType);

    xTrace1(floodp, TR_MORE_EVENTS, "floodOpenDisable returning %x", val);
    return val;
}

/*
 * floodOpenDone: Upon notification from the lower protocol that a session
 * has been open, it must get the participant list for the new session in
 * order to determine if it is using a flooding address or not, and set
 * the flag appropriately.
 */
static xkern_return_t
floodOpenDone(self, lls, llp, hlpType)
    XObj self, lls, llp, hlpType;
{
    XObj        s;
    PState      *pstate;
    Enable      *e;
   
    xTrace0(floodp, TR_FULL_TRACE, "Entered floodOpenDone");

    if ( self == hlpType ) {
        xTrace0(floodp, TR_ERRORS, "self == hlpType in floodOpenDone");
        return XK_FAILURE;
    }
    pstate = (PState *)self->state;

    /* check for open enables pending for the higher protocol.
     */
    if ( mapResolve(pstate->pasMap, &hlpType, &e) == XK_FAILURE ) {
        /* This shouldn't happen, since a pending open enable for flood
	 * has to correspond to a higher protocol's same operation, but...
         */
        xTrace0(floodp, TR_ERRORS,
                "floodOpenDone: Couldn't find hlp for incoming session");
        return XK_FAILURE;
    }

    xDuplicate(lls); /* Hmm... Not sure about this; it was used in confounder*/

    /* Everything checked, create a new session, and set flags if needed
     */
    if ( (s=floodCreateSessn(self, e->hlpRcv, e->hlpType, lls)) == ERR_XOBJ )
    {
        xClose(lls);
        return XK_FAILURE;
    }
    else
    {
      IPhost destn;

      /* Here we use the GETMYHOST control operation to determine what was
       * the destination address used to reach this host, which may be
       * an interface address, a broadcast address or a flood-specific one.
       * A call to must_flood should determine it.
       */
      xControl(s,GETMYHOST,(void*)&destn,sizeof(destn));
      if ( must_flood(pstate,&destn) )
      {
	((SState*) s->state)->flags |= 2;
	xTrace0(floodp,TR_FULL_TRACE,
	        "floodOpenDone: session marked for flooding");
      }
    }

    xTrace0(floodp, TR_EVENTS,
            "floodOpenDone: Passively opened session successfully created");
    /* If everything worked here, now it is time to pass the information up.
     */
    return xOpenDone(e->hlpRcv, s, self);
}

/*
 * floodDemux: Determine the session to be used just by checking the
 * active map with the lower level session.
 */
static xkern_return_t
floodDemux( XObj self, XObj lls, Msg *msg )
{
  PState	*ps = (PState *)self->state;
  XObj 	s = 0;
  
  xIfTrace(floodp, TR_DETAILED) msgShow(msg);

  if ( mapResolve(ps->actMap, &lls, &s) == XK_FAILURE ) { 
	xTrace0(floodp, TR_EVENTS, "floodDemux: map resolve fails");
  	return XK_FAILURE;
  }

  if ( ! xIsSession(s) ) {
    xTrace0(floodp, TR_FUNCTIONAL_TRACE, "floodDemux returning XK_FAILURE");
    return XK_FAILURE;
  }

  return xPop(s, lls, msg, 0);
}

/*
 * floodClose: Destroy session, close lower session, make sure the binding
 * is removed, assert everything is ok.
 */
static xkern_return_t
floodClose( s )
    XObj	s;
{
    PState	*ps = (PState *)xMyProtl(s)->state;
    xkern_return_t res;

    xTrace1(floodp, TR_FULL_TRACE, "floodClose: closing session %x", s);
    xAssert( xIsSession( s ) );
    xAssert( s->rcnt == 0 );
    res = mapRemoveBinding( ps->actMap, s->binding );
    xAssert( res != XK_FAILURE );
    xClose(xGetDown(s,0));
    xDestroy( s );
    xTrace0(floodp, TR_DETAILED, "floodClose returning");
    return XK_SUCCESS;
}

/* floodPush: If session requires flooding, push the current flood stamp
 * on it and update the stamp identifier, and finally push it to all 
 * interfaces. In any case, push the message down the interface opened
 * by the higher layer.
 */
static xmsg_handle_t
floodPush( XObj	s, Msg 	*msg )
{
  SState* ss = (SState*)(s->state);
  PState* ps = (PState *)(xMyProtl(s)->state);

  xTrace1(floodp, TR_MAJOR_EVENTS, "floodPush: msg size is %d", msgLen(msg));
  xAssert(xIsSession(s));

  if ( ss->flags & 2 ) {
    /* Since this is a flooding session, push the flood stamp to it
     */
    msgPush(msg,flood_hdr_store,&(ps->next_stamp),FLOOD_LEN,0);
    xTrace3(floodp, TR_MAJOR_EVENTS, "floodPush: added stamp (%s,%d), sz %d",
	    ipHostStr(&(ps->next_stamp.host_addr)), ps->next_stamp.seqn,
	    msgLen(msg));
	    
    /* Now add the stamp to the log of seen stamps, to avoid this message
     * get re-flooded later, and update the stamp for the next time.
     */
    flood_log_add( ps, &(ps->next_stamp) );
    generate_new_stamp(ps);

    /* 
     * Earlier flood was sending the message on all interfaces
     * other than the one identified in the existing lower session.
     * Now that we are using class D addresses, vnet multicast will take 
     * care of that. So we only need to do the single push.
     */
  }
  else
  {
    xTrace0(floodp, TR_MAJOR_EVENTS, "floodPush: no stamp added");
  }

  xTrace1(floodp,TR_MAJOR_EVENTS,"floodPush: pushing msg(%d)",msgLen(msg));
  /* 
   * Whether this is a flooding session or not we do this. If it is
   * not a flooding session this does not change anything. If it is
   * Then a class D address will be used and the message will be sent 
   * on all interfaces by vnet multicast 
   */
  return xPush(xGetDown(s, 0), msg);
}

/*
 * floodPop: If this is a flooding message, it must first verify if the
 * newly received message has not been flooded by it before. If that's the
 * case, the message is simply discarded. If it's a new message it
 * verifies if any IP pseudo-header adjustments are needed.
 * In any case, it passes the message up to the higher protocol.
 */
static xkern_return_t
floodPop( XObj s, XObj lls, Msg* msg, void* hdr)
{
  long len;
  flood_stamp_t msg_stamp;
  SState* ss = (SState*)(s->state);
  PState* ps = (PState *)(xMyProtl(s)->state);
  Msg original_msg;
  xTrace1(floodp, TR_MAJOR_EVENTS, "floodPop: msg len %d",msgLen(msg) );

  len = msgLen(msg);

  if ( is_flood_session(ss) ) {
    /* The rest of the action takes place only in this case...
     */
    if (len<FLOOD_LEN)
    {
      xTrace2(floodp, TR_ERRORS,
	      "floodPop received bad message, length %D less than %d",
	      len, FLOOD_LEN);
      return XK_FAILURE;
    }
  /*
   * Should it build a copy if the message to be handled on each retransmission?
   */

    /* Remove the flood stamp to determine if message is new or not.
     * Before that, keep a copy of the original message in case we have to
     * push it down again.
     */
    msgConstructCopy(&original_msg,msg);
    msgPop(msg,flood_hdr_load,&msg_stamp,FLOOD_LEN,0);

    xTrace3(floodp, TR_MAJOR_EVENTS,"floodPop: removing stamp (%s,%d) sz %d",
	    ipHostStr(&(msg_stamp.host_addr)), msg_stamp.seqn,
	    msgLen(msg));

    /* Decide if message must be flooded to other interfaces or discarded
     */
    if ( flood_log_find( ps, &msg_stamp ) ) {
      xTrace0(floodp,TR_MAJOR_EVENTS,"floodPop: msg already in log, discarded");
      msgDestroy(&original_msg);
      return XK_SUCCESS;
    }
    else {
      /* Message has not been seen before. It must first fix the IP pseudoheader
       * if necessary, and then copy the message down again (with the original
       * stamp) to all interfaces other than the one it was received from.
       */
      if ( ss->flags & 1 ||((PState *)(xMyProtl(s)->state))->flags & 1 )
      {
         /* IP pseudo-header fixup code, from confounder 
	  */
         IPpseudoHdr *ipphdr = msgGetAttr( msg, 0);
  
         if (ipphdr) {
  	   ipphdr->len = htons(ntohs(ipphdr->len)-FLOOD_LEN);
  	   xTrace3(floodp, TR_MAJOR_EVENTS,
  		  "floodPop:  Adjusting IP pseudoheader length; "
  		  "ipphdr %x  phlen after %d  len before %d",
  		  ipphdr, ntohs(ipphdr->len), len);
         }
         else {
  	   xTrace0(floodp, TR_MAJOR_EVENTS,
  	   "floodPop: Skipping length adjustment for IP pseudoheader ");
         };
      }
  
      xTrace0(floodp,TR_MAJOR_EVENTS,"floodPop: new msg");

      /* Now to flood specific actions. First, add stamp to the log, so
       * this message will be discarded if received again.
       * Then proceed to determine what was the participant list used
       * on openning this session, and the interface from which the message
       * was received.
       */
      flood_log_add( ps, &msg_stamp );

      {
	IPhost  destn, *interface;
	Part    p[3];
	char    buffer[2048];

	/* GETMYHOST returns the destination address used by the sender
	 * to identify this session. This is the address we must use to
	 * open the new sessions.
	 */
	xControl(s,GETMYHOST,(void*)&destn,sizeof(destn));

        /* GETPARTICIPANTS returns the participant list used. We can
	 * pop/push back the local part to determine the interface through
	 * which the message was received, and replace the remote part
	 * (initially the sender's address) by the destination address
	 * obtained with GETMYHOST.
	 */
	xControl(s,GETPARTICIPANTS,buffer,sizeof(buffer));
	partInternalize(p,buffer);
        interface = partPop(p[1]);
	partPush(p[1],interface,sizeof(IPhost));
	partPop(p[0]);                         /* removes address of sender */
	partPush(p[0],(void*)&destn,sizeof(destn)); /* Adds original destn. */

	/* Now we have the participant list that would have been used to send
	 * the message down the interface it came, so we can pass that to
	 * do_flood for it to use to determine the participant lists to be
	 * used when pushing the message down to other interfaces.
	 */
        lls = xOpen(xMyProtl(s), xHlpType(s), xGetDown(xMyProtl(s), 0), p);
        if ( !xIsSession(lls) ) {
            xTrace0(floodp,TR_ERRORS,"floodPop: unable to open session");
        }
        xPush(lls,&original_msg);
        xClose(lls);
	msgDestroy(&original_msg);
      }
    }
  }
  else {
    xTrace0(floodp, TR_MAJOR_EVENTS,"floodPop: no stamp removed");
  }

  /* Finally, pass the message up.
   */
  return xDemux(s, msg);    
} 

/* floodControlSessn: The control operations here came from confounder. They
 * are used to fix the IP pseudo-header if needed. In this case we added
 * tests to verify if the session uses flooding, since only in this case
 * the message length is altered by flood.
 */
static int
floodControlSessn(s, op, buf, len)
    XObj	s;
    int 	op, len;
    char 	*buf;
{
    xAssert(xIsSession(s));
    switch (op) {

      case GETMAXPACKET:
      case GETOPTPACKET:
	{ 
	    int max;
	    if (len < sizeof(max) || !buf) return 0; 
	    /*
	     * First get the information from lower layers.
	     */
	    if (xControl(xGetDown(s, 0), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
	    /*
	     * Now, if this session do flooding (what will change the message
	     * length), change the packet size information passed back.
	     */
	    if (is_flood_session(s->state))
	      max -= FLOOD_LEN;
	    if (max<0) return 0;
	    *(int *)buf = max;
	    return(sizeof(max));
	}

      case IP_PSEUDOHDR:
      /* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((SState *)(s->state))->flags |= 1;
	    xTrace0(floodp, TR_EVENTS,
	       "floodControlSessn: Turning on IPpseudoheader length fixup.");  }
      /***** FALL THROUGH TO DEFAULT CASE, so lower sessions see it too. *****/

      default:
	return xControl(xGetDown(s,0), op, buf, len);
    }
}

/* floodControlProtl: Just as floodControlSessn, this code comes from
 * confounder, with little changes.
 */
static int
floodControlProtl( self, op, buf, len )
    XObj	self;
    int 	op, len;
    char 	*buf;
{
    xAssert(xIsProtocol(self));

    switch (op) {
      case GETMAXPACKET:
      case GETOPTPACKET:
        {
            int max;
	    if (len < sizeof(max) || !buf) return 0; 
	    if (xControl(xGetDown(self,0),op,(char*)&max,sizeof(max)) <= 0) {
		return 0; 
	    }
	    /*
	     * In this case, for the protocol as a whole, we are not able to
	     * determine if the message will be changed or not. The solution
	     * right now is to take the worst case, and reduce the message size.
	     */
	    max -= FLOOD_LEN;
	    if (max<0)
	      return 0;
            *(int *)buf = max;
            return(sizeof(max));
	}

      case IP_PSEUDOHDR:
      /* Turn on flag saying to fixup length on IPpseudoheader.
       */
	{   ((PState *)(self->state))->flags |= 1;
	    xTrace0(floodp, TR_EVENTS,
	       "floodControlProtl: Turning on IPpseudoheader length fixup.");  }
      /***** FALL THROUGH TO DEFAULT CASE, so lower protocols see it too. *****/

      default:
	return xControl(xGetDown(self, 0), op, buf, len);
  }
  
}

/* floodSessnInit: Just initialize the xkernel functions. Session state
 * must be initialized by the caller.
 */
static void
floodSessnInit(s)
    XObj s;
{
  s->push = floodPush;
  s->pop = floodPop;
  s->close = floodClose;
  s->control = floodControlSessn;
}

/* AQUI */

/*
 * flood_log_init: flood_log is kept in the protocol state in order to record
 * which messages were seen in the near past. It is implemented as a timing
 * wheel, where the "wheel" is an array of mappings holding the stamps of
 * messages seen. The initialization is just a matter of creating the
 * mappings.
 */
void
flood_log_init( XObj protl )
{
  int i;
  PState* ps = (PState*)(protl->state);

  for ( i=0;i<FLOOD_LOG_SZ;++i )
    PS_MSG_LOG_IDX(ps,i) = mapCreate( 2*FLOOD_LOG_RATE, sizeof(flood_stamp_t) );
}

/*
 * flood_log_add: just binds a new stamp to the current slot of the wheel.
 */

void
flood_log_add( PState* ps, flood_stamp_t* stamp )
{
  if ( mapBind( PS_CURR_MSG_LOG(ps), stamp, NULL ) == ERR_BIND ) {
    xTrace1(floodp,TR_ERRORS,"flood: unable to add stamp to log at slot %d",
            PS_CURR_SLOT(ps));
  }
}

/*
 * Flood log find must check all the slots of the wheel, not just the
 * current one, to see if the given stamp was logged to any position.
 */

int
flood_log_find( PState* ps, flood_stamp_t* stamp )
{
  int i;

  for (i=0;i<FLOOD_LOG_SZ; ++i) {
    if ( mapResolve( PS_MSG_LOG_IDX(ps,i), stamp, NULL) == XK_SUCCESS )
      return 1;
  }
  return 0;
}

/*
 * flood_log_update: It's fired as an event, periodically. It's actions
 * include updating the current slot identifier, which then has its
 * mapping emptied (closed and created again) to discard logged stamps
 * that now have become too old. At the end, it schedules itself again
 * to be executed after another interval.
 */

void
flood_log_update( Event ev, void* arg )
{
  XObj protl = (XObj) arg;
  PState* ps = (PState*)(protl->state);

  PS_NEXT_SLOT(ps);
  xTrace1(floodp,TR_FULL_TRACE,"flood_log_update executing on slot %d",
          PS_CURR_SLOT(ps));
  mapClose( PS_CURR_MSG_LOG(ps) );
  PS_CURR_MSG_LOG(ps) = mapCreate( 2*FLOOD_LOG_RATE, sizeof(flood_stamp_t) );
  ev = evSchedule( flood_log_update, (void*)protl, FLOOD_LOG_INTERVAL*1000000 );
  evDetach( ev );
}

/*
 * flood_hdr_store and flood_hdr_load: the typical store/load header functions,
 * make use of the usc generated functions/macros defined in "floodhdr.h".
 */

void
flood_hdr_store( VOID* hdr, char* dst, long  len, VOID* arg )
{
  xTrace0(floodp, TR_EVENTS, "flood_hdr_store" );

  if ( ( ((int)dst) % 4 ) == 0 ) {
    native2net4( hdr, dst );
  }
  else {
    native2net( hdr, dst );
  }
}

long
flood_hdr_load( VOID* hdr, char* src, long len, VOID* arg)
{
  xTrace0(floodp, TR_EVENTS, "flood_hdr_load" );

  if ( ( ((int)src) % 4 ) == 0 ) {
    net42native( src, hdr );
  }
  else {
    net2native( src, hdr );
  }
  return len;
}
