/*
 * $RCSfile: grpmemtst.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Log: grpmemtst.c,v $
 * Revision 1.1  1995/04/24 18:48:15  dorgival
 * Initial revision
 *
 * Revision 1.1  1995/04/05  17:52:37  dorgival
 * Initial revision
 *
 *
 */

/*
 * grpmemtst.c
 *
 * Simple test file for grpmem. It should work as a protocol on top
 * of grpmem issuing opens, closes and control ops on grpmem and its
 * sessions.
 */

#include <stdio.h>
#include "xkernel.h"
#include "km.h"
#include "grpmem.h"
#include "host_list.h"
#include "grpmem_i.h"

#define isspace(c) (((c)==' ')||((c)=='\n')||((c)=='\t')||((c)=='\0'))

int tracegrpmemtstp;
XObj Self;

void grpmemtst_init( XObj );
char* hosts2str( void );
void str2hosts( char* str );
XObj open_group( IDTYPE gid, XObj self );
void del_group( IDTYPE gid );
void add_members( XObj grp, char* members );
void del_members( XObj grp, char* members );
void list_all_groups( void );
void list_all_members( XObj grp );
void test_event( void );

XObj   grpmem;
XObj   sessn;
IDTYPE groups[128];
LOCALHOSTTYPE hosts[128];

char*
hosts2str( void )
{
  static char buf[1024];
  int    i;
  buf[0] = '\0';
  for ( i=0; *((int*)&hosts[i]); ++i ) {
    strcat(buf," ");
    strcat(buf,ipHostStr(&(hosts[i])));
  }
  return buf;
}

void
str2hosts( char* str )
{
  char* cptr = str;
  char* ahost = cptr;
  int   i = 0;

  do {
    if ( isspace(*cptr) ) {
      *cptr = '\0';
      str2ipHost(&(hosts[i++]), ahost);
      ahost = cptr+1;
    }
  } while ( *(cptr++) );
  bzero(&hosts[i],sizeof(LOCALHOSTTYPE));
}

XObj
open_group( IDTYPE gid, XObj self )
{
  Part	p[1];
  XObj  grpsessn;

  partInit(p, 1);

  partPush(p[REMOTE_PART], &gid, sizeof(gid) );

  fprintf(stderr,"grpmemtst: Opening group %d\n",gid);

  grpsessn = xOpen(self,self,grpmem,p);
  if ( grpsessn == ERR_XOBJ ) {
    xError("grpmemtst unable to open session");
  }
  return grpsessn;
}

void
del_group( IDTYPE gid )
{
  fprintf(stderr,"grpmemtst: Deleting group %d\n",gid);
  xControl(grpmem,GRPMEM_DELGRP,(void*)&gid, sizeof(IDTYPE));
}

void
add_members( XObj grp, char* members )
{
  str2hosts( members );
  fprintf(stderr,"Adding %s...", hosts2str() );
  xControl(grp,GRPMEM_ADDMEM,(char*)hosts,sizeof(hosts));
  fprintf(stderr," Done.\n");
}

void
del_members( XObj grp, char* members )
{
  str2hosts(members);
  fprintf(stderr,"grpmemtst: Deleting %s... ",hosts2str());
  xControl(sessn,GRPMEM_DELMEM,(char*)hosts,sizeof(hosts));
  fprintf(stderr,"Done.\n");
}

void
list_all_groups( void )
{
  int i;
  xControl(grpmem,GRPMEM_CNTGRP,(void*)&i,sizeof(i));
  fprintf(stderr,"grpmemtst: CNTGRP = %d\n",i);
  if ( xControl(grpmem,GRPMEM_ALLGRP,(char*)groups,sizeof(groups)) != i ) {
    fprintf(stderr,"grpmemtst: Return value of ALLGRP does not match CNTGRP!\n");
  }
  fprintf(stderr,"grpmemtst: ALLGRP =");
  while ( i-- ) {
    fprintf(stderr," %d", groups[i]);
  }
  fprintf(stderr,"\n");

}

void
list_all_members( XObj grp )
{
  int i;

  fprintf(stderr,"grpmemtst: group members\n");
  xControl(grp,GRPMEM_CNTMEM,(void*)&i,sizeof(i));
  fprintf(stderr,"grpmemtst: CNTMEM = %d\n",i);
  if ( i != xControl(grp,GRPMEM_ALLMEM,(char*)hosts,sizeof(hosts)) ) {
    fprintf(stderr,"grpmemtst: Return value of ALLMEM does not match CNTMEM!\n");
  }
  fprintf(stderr,"grpmemtst: ALLMEM =");
  while ( i-- ) {
    fprintf(stderr," \"%s\"", ipHostStr(&hosts[i]));
  }
  fprintf(stderr,"\n");
}

void
grpmemtst_init( self )
  XObj self;
{
  if ( globalArgc < 2 ) {
    fprintf(stderr,"Not server, returning\n");
    return;
  }
  else {
    Self = self;
    evSchedule((EvFunc)test_event,NULL,10000000);
  }
}

void
test_event( void )
{
  xTrace0(grpmemtstp, TR_GROSS_EVENTS, "GRPMEMTST init");
  xAssert(xIsProtocol(Self));
  grpmem = xGetDown(Self, 0);
  if ( ! xIsProtocol(grpmem) ) {
    xTrace0(grpmemtstp, TR_ERRORS, "GRPMEMTST can't find grpmem");
    xError("GRPMEMTST init errors -- not initializing");
    return;
  } 
  if ( globalArgc == 1 ) {
    xTrace0(grpmemtstp, TR_ALWAYS, "GRPMEMTST returning - client");
    return; /* Only an owner may exist and it is started with "-s" */
  }

  fprintf(stderr,"grpmemtst: Listing all groups read from rom file\n");
  list_all_groups();

  sessn = open_group(1001, Self);

  list_all_members(sessn);

  add_members(sessn, "192.12.69.4");

  list_all_members(sessn);

  del_members(sessn,"192.12.69.2");

  list_all_members(sessn);

  del_members(sessn,"192.12.69.3");

  list_all_members(sessn);

  add_members(sessn, "192.12.69.2");

  list_all_members(sessn);

  xClose(sessn);
}

