/*
 * $RCSfile: grpmem.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Log: grpmem.c,v $
 * Revision 1.2  1996/04/10 00:33:47  rajesh
 * change to use Class D addresses. Changed km_update.
 *
 * Revision 1.2  1995/04/28  00:21:08  dorgival
 * Minor syntax changes to make the IRIX compiler happy
 *
 * Revision 1.1  1995/04/24  18:48:15  dorgival
 * Initial revision
 *
 * Revision 1.1  1995/04/05  17:52:37  dorgival
 * Initial revision
 *
 *
 */

/*
 *   GROUP MEMBERSHIP PROTOCOL
 *
 *   First version:
 *
 *   Fixed owner, only one to make changes to group
 *   Group list is distributed periodically
 *   Members that timeout waiting for updates remove themselves from group
 *   No checks about identity of sender
 */

#include "xkernel.h"
#include "km.h"
#include "grpmem.h"
#include "host_list.h"
#include "grpmem_i.h"

#define USE_BRDCST 1

#define savePart(pxx, spxx)	\
    bcopy((char *)(pxx), (char *)(spxx), sizeof(Part) * partLen(p))

int tracegrpmemp;

LOCALHOSTTYPE null_addr = { 0,0,0,0 };

#define addr_is_valid(host_addr)  ((char*)bcmp((char*)host_addr,(char*)&null_addr,sizeof(null_addr)))

#ifdef __STDC__

void	grpmem_init( XObj );

static LOCALHOSTTYPE* bcast_addr( LOCALHOSTTYPE* interface );

static xkern_return_t   grpmemClose( XObj sessn );
static XObj		grpmemOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	grpmemOpenDone( XObj, XObj, XObj, XObj );
static xkern_return_t	grpmemOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	grpmemOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t   grpmemClose( XObj sessn );
static int		grpmemControlProtl( XObj, int, char *, int );
static int              grpmemControlSessn( XObj, int, char*, int );
static SState*          grpmemCreateSState( XObj );
static XObj             grpmemCreateSessn( XObj, XObj, XObj,
					   IDTYPE,int,LOCALHOSTTYPE*);
static void             grpmemDumpSessn( SState* );
static void		getproc_sessn( XObj );
static xkern_return_t   romrule( XObj, char **, int, int, void * );
static xkern_return_t   romport( XObj, char **, int, int, void * );
static xkern_return_t   romupdate( XObj, char **, int, int, void * );
static xkern_return_t   romaddress( XObj, char **, int, int, void * );

static grpmem_cmd_codes cmd_to_int(char *);

static xkern_return_t grpmemDemux( XObj self, XObj s, Msg* msg );
static xkern_return_t grpmemPop( XObj self,XObj lls, Msg* msg, grpmem_msg* hdr);

static int foreach_cntmem( LOCALHOSTTYPE* ahost, int* cntmem );
static int foreach_allmem( LOCALHOSTTYPE* ahost, LOCALHOSTTYPE** buf );

static int map_foreach_cntgrp( IDTYPE* grpid, XObj grpsessn, int* cntgrp );
static int map_foreach_allgrp( IDTYPE* grpid, XObj grpsessn, IDTYPE** allgrp );

static void global_delgrp( XObj protl, XObj sessn );

static void local_delgrp( XObj sessn );
static int  local_cntgrp( XObj self );
static void local_allgrp( XObj   self, IDTYPE *buf );

static void grpmem_update( SState*, int cnt, LOCALHOSTTYPE* hosts);
static void grpmem_addmem( SState*, LOCALHOSTTYPE* hosts);
static void grpmem_delmem( SState*, LOCALHOSTTYPE* hosts);
static int  grpmem_cntmem( SState* );
static int  grpmem_allmem( SState*, LOCALHOSTTYPE* buf, int len);

static void km_update( SState* );
static void km_invalidate( SState* ss );

static void update_event( Event e, XObj protl );
static int  periodic_group_check( IDTYPE* gidptr, XObj sessn );

static void send_update( SState* );
static void send_query( SState* );

static void dump_grpmem_msg( grpmem_msg* msg );

static grpmem_msg* create_group_description( grpmem_cmd_codes cmd, SState* ss );
static Msg*        build_delgrp_msg( SState* ss );

static void set_part_send( Part* p, LOCALHOSTTYPE* host, PState* ps );
static void set_part_recv( Part* p, PState* ps );

static int grpmem_send_msg( LOCALHOSTTYPE* host, Msg* orig, SState* ss );
static int grpmem_bcast_msg( LOCALHOSTTYPE* host, Msg* orig, SState* ss );
static int its_myself( PState* ps, LOCALHOSTTYPE* host );

/* The next two are in the usit/usc source file */

void grpmem_hdr_store( VOID* hdr, char* dst, long  len, VOID* arg );
long grpmem_hdr_load(  VOID* hdr, char* dst, long  len, VOID* arg );

#endif

static LOCALHOSTTYPE*
bcast_addr( LOCALHOSTTYPE* interface )
{
  static LOCALHOSTTYPE bcast;
  LOCALHOSTTYPE mask, cmask;

  netMaskFind( &mask, interface );
  IP_COMP( cmask, mask );
  IP_OR( bcast, cmask, *interface );

  return &bcast;
  /*
  { static IPhost classD; str2ipHost(&classD,"224.1.1.1"); return &classD; }
  */
}

static int
its_myself( PState* ps, LOCALHOSTTYPE* host )
{
  int i, interface_cnt = PS_IF_CNT(ps);
  for ( i=0; i<interface_cnt; ++i ) {
    if ( same_host( host, PS_INTERFACE(ps,i) ) ) {
      break;
    }
  }
  return ( i < interface_cnt );
}

static void
dump_grpmem_msg( msg )
  grpmem_msg* msg;
{
  int i;
  fprintf(stderr,"\tgrpmem_msg: op = %d, gid = %d, host count = %d\n",
		  MSG_COM(msg), MSG_GID(msg), MSG_HOST_COUNT(msg) );
  fprintf(stderr,"\tgrpmem_msg hosts:");
  for (i=0;i<MSG_HOST_COUNT(msg);++i) {
    fprintf(stderr," %s",ipHostStr(MSG_HOST(msg,i)) );
  }
  fprintf(stderr,"\n");
}

static grpmem_msg*
create_group_description( grpmem_cmd_codes cmd, SState* ss )
{
  int i,j,count;
  grpmem_msg* desc;

  xTrace0(grpmemp, TR_MAJOR_EVENTS, "create_msg");
  /*
  count = SS_GROUP_OWNER_CNT(ss) + SS_GROUP_ADMIN_CNT(ss) + SS_GROUP_SIZE(ss)+2;
  */
  count = SS_GROUP_OWNER_CNT(ss) + SS_GROUP_SIZE(ss)+1;

  desc = MSG_NEW(count);
  MSG_COM(desc)        = cmd;
  MSG_GID(desc)        = SS_GROUP_ID(ss);
  MSG_HOST_COUNT(desc) = count;

  for (i=0;i<SS_GROUP_OWNER_CNT(ss);++i) {
    bcopy((char*)SS_GROUP_OWNER_ADDR(ss,i), (char*)MSG_HOST(desc,i), sizeof(LOCALHOSTTYPE));
  }
  bzero((char*)MSG_HOST(desc,i), sizeof(LOCALHOSTTYPE));

  /*
  for (j=0,++i;j<SS_GROUP_ADMIN_CNT(ss);++j,++i) {
    bcopy((char*)SS_GROUP_ADMIN(ss,j), (char*)MSG_HOST(desc,i), sizeof(LOCALHOSTTYPE));
  }
  bzero((char*)MSG_HOST(desc,i), sizeof(LOCALHOSTTYPE));
  */

  for (j=0,++i;j<SS_GROUP_SIZE(ss);++j,++i) {
    bcopy((char*)SS_GROUP_MEMBER(ss,j),(char*)MSG_HOST(desc,i),sizeof(LOCALHOSTTYPE));
  }

  xIfTrace(grpmemp,TR_FULL_TRACE)
    dump_grpmem_msg(desc);
  
  return desc;
}


static Msg*
build_update_msg( SState* ss )
{
  int count;
  grpmem_msg* msg_native;
  static Msg msg_net;

  xTrace0(grpmemp, TR_MAJOR_EVENTS, "build_update_msg");

  msg_native = create_group_description( GRP_UPDATE, ss );
  count = MSG_HOST_COUNT(msg_native);

  msgConstructEmpty( &msg_net );
  msgPush( &msg_net, grpmem_hdr_store, msg_native, MSG_SIZE(count), NULL );
  xFree( (char*) msg_native );

  return &msg_net;
}

static Msg*
build_delgrp_msg( SState* ss )
{
  grpmem_msg* msg_native;
  static Msg msg_net;
  
  xTrace0(grpmemp, TR_MAJOR_EVENTS, "build_delgrp_msg");
  msg_native = MSG_NEW(0);
  MSG_COM(msg_native)        = GRP_DEL;
  MSG_GID(msg_native)        = SS_GROUP_ID(ss);
  MSG_HOST_COUNT(msg_native) = 0;

  xIfTrace(grpmemp,TR_FULL_TRACE)
    dump_grpmem_msg(msg_native);

  msgConstructEmpty( &msg_net );
  msgPush( &msg_net, grpmem_hdr_store, msg_native, MSG_SIZE(0), NULL );
  xFree((char*) msg_native);
  
  return &msg_net;
}

static Msg*
build_request_msg( SState* ss )
{
  grpmem_msg* msg_native;
  static Msg msg_net;
  
  xTrace0(grpmemp, TR_MAJOR_EVENTS, "build_request_msg");
  msg_native = MSG_NEW(0);
  MSG_COM(msg_native)        = GRP_REQUEST;
  MSG_GID(msg_native)        = SS_GROUP_ID(ss);
  MSG_HOST_COUNT(msg_native) = 0;

  xIfTrace(grpmemp,TR_FULL_TRACE)
    dump_grpmem_msg(msg_native);

  msgConstructEmpty( &msg_net );
  msgPush( &msg_net, grpmem_hdr_store, msg_native, MSG_SIZE(0), NULL );
  xFree((char*) msg_native);
  
  return &msg_net;
}

static void
set_part_send( p, host, ps )
  Part* p;
  LOCALHOSTTYPE* host;
  PState* ps;
{
  static long local_port, remote_port;

  local_port  = ANY_PORT;
  remote_port = PS_CNTRL_PORT( ps );
  partInit(p,2);
  partPush(p[0],host,sizeof(LOCALHOSTTYPE));
  partPush(p[0],&remote_port,sizeof(long));
  partPush(p[1],ps->my_host,sizeof(LOCALHOSTTYPE));
  partPush(p[1],&local_port,sizeof(long));
}

static void
set_part_recv( p, ps )
  Part* p;
  PState* ps;
{
  static long local_port;

  local_port = PS_CNTRL_PORT( ps );

  partInit(p, 2);
  partPush(p[0], ANY_HOST, 0);
  partPush(p[0], &local_port, sizeof(long));
  /*
  partPush(p[1],ANY_HOST,0);
  partPush(p[1],&local_port,sizeof(long));
  */
}

static xkern_return_t
grpmemClose( XObj sessn )
{
  SState* ss = (SState*)sessn->state;
  PState* ps = (PState*)SS_GROUP_PROTL(ss)->state;
  IDTYPE  id = SS_GROUP_ID(ss);
  XObj    tmp;

  if ( mapResolve(ps->group_map,(void*)&id,(void**)&tmp)
       == XK_FAILURE ) {
    xTrace1(grpmemp,TR_MAJOR_EVENTS,"grpmemClose: group %d",SS_GROUP_ID(ss));
    xDestroy( sessn );
  }
  return XK_SUCCESS;
}

/*
 *  grpmemOpen
 *
 *    The remote part is the group name; we will return a session to
 *    use for updates to the group list.  Updates will be sent to
 *    all hosts on the list.
 *
 */

static XObj
grpmemOpen( self, hlpRcv, hlpType, p )
  XObj	self, hlpRcv, hlpType;
  Part 	*p;
{
  SState*      ss;
  PState*      ps = (PState*)(self->state);
  IDTYPE       group_id;
  LOCALHOSTTYPE* hosts;
  XObj	       sessn;
  int          host_cnt = 0;
  int          interface_cnt = PS_IF_CNT(ps);
  
  xTrace0(grpmemp, TR_MAJOR_EVENTS, "grpmemOpen");

  group_id = *((IDTYPE*)partPop(p[0]));

  xTrace3(grpmemp,TR_FULL_TRACE,"grpmemOpen group %d, %d owner addrs, %d hosts",
		  group_id, interface_cnt, host_cnt);

  if(mapResolve(ps->group_map,(void*)&group_id,(void**)&sessn)
     !=XK_SUCCESS)
  {
    xTrace0(grpmemp,TR_FULL_TRACE, "grpmemOpen: creating new group");
    sessn = grpmemCreateSessn( self, hlpRcv, hlpType, group_id, 0, NULL );
    ss    = (SState*) sessn->state;
    km_update( ss );
  }
  else if ( hosts ) {
    xTrace1(grpmemp, TR_ERRORS,
	    "grpmemOpen: should not provide elements for existing group %d",
	    group_id);
  }

  return sessn;
}

/*
 *  grpmemOpenEnable
 *
 *   This routine can be used by user interface protocols that want
 *    notifications when a new group is created.
 *
 *    assume there is only one participant, the convention is that it specifies
 *    the local part ... e.g. the local server port for this machine.
 *
 */

static xkern_return_t
grpmemOpenEnable( self, hlpRcv, hlpType, p )
  XObj	self, hlpRcv, hlpType;
  Part	*p;
{
  signed IDTYPE* hostport = partPop(p[REMOTE]);  /* the name of the group */
  PState*        ps = (PState *)self->state;
  Enable*        e;

  xTrace0(grpmemp, TR_MAJOR_EVENTS, "grpmemOpenEnable");

  if (mapResolve(ps->passive,hostport,(void*)e)!=XK_FAILURE)
  {
    if ( e->hlpRcv == hlpRcv ) {
	e->rcnt++;
	return XK_SUCCESS;
    }
    return XK_FAILURE;
  }
  e = (Enable *)xMalloc(sizeof(Enable));
  e->hlpRcv = hlpRcv;
  e->hlpType = hlpType;
  e->rcnt = 1;
  e->binding = mapBind(ps->passive, hostport, e);
  if (e->binding == ERR_BIND) {
    xFree((char *)e);
    return XK_FAILURE;
  }
  return XK_SUCCESS;
}

static xkern_return_t
grpmemOpenDisable( self, hlpRcv, hlpType, p )
  XObj self, hlpRcv, hlpType;
  Part *p;
{
  signed IDTYPE* hostport = partPop(p[REMOTE]);  /* the name of the group */
  PState*       ps = (PState *)self->state;
  Enable*       e;

  xTrace0(grpmemp, TR_MAJOR_EVENTS, "grpmemOpenDisable");
  if ( ( mapResolve(ps->passive,hostport,(void*)&e)
       !=XK_FAILURE)
     &&(e->hlpRcv == hlpRcv)&&(e->hlpType == hlpType))
  {
    if (--(e->rcnt) == 0) {
	mapRemoveBinding(ps->passive, e->binding);
	xFree((char *)e);
    }
    return XK_SUCCESS;
  } else {
    return XK_FAILURE;
  }
}

/*
 *  grpmemOpenDone
 *
 *     This simply means a new host is
 *	reporting information; we don't necessarily need to open
 *	a session.
 */

static xkern_return_t
grpmemOpenDone( self, lls, llp, hlpType )
  XObj self, lls, llp, hlpType;
{
/*
  PState	*ps = (PState *)self->state;
  Enable	*e;
  XObj	lkp = ERR_XOBJ;
  Part	p[2], sp[2];
  char	partbuf[sizeof(p)], *partbuf2;
  xkern_return_t	val;
  Part	*partlist;
  int		bufsz;
*/
  
  xTrace0(grpmemp, TR_MAJOR_EVENTS, "grpmemOpenDone");
  return XK_SUCCESS;
}


/*
 * grpmemDemux
 *
 *   incoming message; find the session for the group name
 *    or create a new one
 *
 */

static xkern_return_t
grpmemDemux(protl, lls, msg)
  XObj protl;
  XObj lls;
  Msg* msg;
{
  PState*          ps;
  SState*          ss;
  grpmem_msg*      hdr;
  IDTYPE           id;
  LOCALHOSTTYPE    sender;
  XObj             ssn = ERR_XOBJ;

  xTrace0(grpmemp, TR_MAJOR_EVENTS, "grpmemDemux");

  ps = (PState*)protl->state;

  hdr = MSG_NEW(GRPMEM_MAX_PARTS);
  xControl(lls, GETPEERHOST, (void*)&sender, sizeof(LOCALHOSTTYPE));
  msgPop(msg, grpmem_hdr_load, hdr, sizeof(grpmem_msg), 0);
  id = hdr->group_id;
  xTrace2(grpmemp, TR_FULL_TRACE, "grpmemDemux: group %d, op %d", id, hdr->cmd );

  if ( its_myself( ps, &sender ) ) {
    xTrace0(grpmemp, TR_FULL_TRACE, "grpmemDemux returning: message from myself");
    xFree((char*)hdr);
    return XK_SUCCESS;
  }

  if ( ( mapResolve(ps->group_map,(void*)&id,(void**)&ssn)==XK_FAILURE)
     &&( hdr->cmd == GRP_UPDATE ) )
  {
    Enable        *e, no_binding;

    xTrace1(grpmemp,TR_FULL_TRACE,"grpmemDemux: adding new group %d",id);

    if (mapResolve(ps->passive,(void*)&id,(void**)&e )
       !=XK_SUCCESS) {
      e = &no_binding; e->hlpRcv = NULL; e->hlpType = NULL;
    }

    ssn = grpmemCreateSessn( protl, e->hlpRcv, e->hlpType,id,0,NULL );
    if ( e != &no_binding ) {
      if (--(e->rcnt) == 0) {
	  mapRemoveBinding(ps->passive, e->binding);
	  xFree((char *)e);
      }
      xOpenDone(e->hlpRcv,ssn,protl);
    }
  }
  ss = (SState*)ssn->state;

  if ( ssn == ERR_XOBJ ) {
    xFree((char*)hdr);
    return XK_FAILURE;
  }
  switch( hdr->cmd ) {

    case GRP_DEL: {
      xTrace0(grpmemp,TR_FULL_TRACE,"grpmemDemux: GRP_DEL");
      local_delgrp( ssn );
      break;
    }

    default: {
      xTrace1(grpmemp,TR_FULL_TRACE,"grpmemDemux: cmd = %d, calling grpmemPop",
		      hdr->cmd);
      xPop( ssn, lls, msg, hdr);
    }
  }
  xFree((char*)hdr);
  return XK_SUCCESS;
}

/*
 * grpmemPop
 *
 *  incoming message; change some aspect of the group
 *
 */

static xkern_return_t
grpmemPop( self, lls, xmsg, msg )
  XObj        self;
  XObj        lls;
  Msg*        xmsg;
  grpmem_msg* msg;
{
  SState*        ss = (SState*)self->state;
  LOCALHOSTTYPE* hosts = MSG_HOSTS(msg);
  int            host_cnt = MSG_HOST_COUNT(msg);

  xTrace0(grpmemp, TR_MAJOR_EVENTS, "grpmemPop");

  switch (msg->cmd) {

    case GRP_UPDATE: {
      grpmem_update( ss, host_cnt, hosts );
      break;
    }

    case GRP_REQUEST: {
      /*
       * Currently there's no restriction about this: any host can answer
       * a request from any other host.
       */
      Msg* reply_msg;
      LOCALHOSTTYPE sender;

      xControl(lls, GETPEERHOST, (void*)&sender, sizeof(LOCALHOSTTYPE));

      xTrace1(grpmemp,TR_FULL_TRACE,"grpmemPop: update REQUEST from %s",
	      ipHostStr(&sender) );
      
      reply_msg = build_update_msg( ss );
      grpmem_send_msg( &sender, reply_msg, ss );
      msgDestroy(reply_msg);
      break;
    }

    default:
	  xTrace1(grpmemp,TR_ALWAYS,"grpmemPop: unexpected command: %d",
		  msg->cmd);
      break;
  }

  return XK_SUCCESS;
}

static XObj
grpmemCreateSessn( protl, hlpRcv, hlpType, group_id, host_cnt, hosts )
  XObj protl, hlpRcv, hlpType;
  IDTYPE group_id;
  int host_cnt;
  LOCALHOSTTYPE* hosts;
{
  PState*      ps = (PState *)(protl->state);
  SState*      ss;
  int          p_total;
  XObj*        p_list;
  XObj	       sessn;

  xTrace1(grpmemp, TR_MAJOR_EVENTS, "grpmemCreateSessn, group %d",group_id);

  p_total = GRPMEM_MAX_PARTS;
  p_list  = (XObj*) xMalloc( p_total * sizeof(XObj) );
  sessn   = xCreateSessn(getproc_sessn,hlpRcv,hlpType,protl,p_total,p_list );
  sessn->rcnt = 1;
  ss      = (SState*) sessn->state;
  mapBind(ps->group_map,(void*)&group_id,sessn);

  SS_GROUP_ID(ss) = group_id;
  if ( host_cnt > 0 )
    grpmem_update( ss, host_cnt, hosts );
  else
    grpmem_update( ss, PS_IF_CNT(ps), PS_INTERFACES(ps) );
  SS_NEXT_QUERY(ss)        = QUERY_PERIOD+1;
  SS_REMAINING_QUERIES(ss) = MAX_QUERIES+1;

  return sessn;
}

static void
grpmemDumpSessn( SState* ss )
{
  int i, max;
  fprintf(stderr,"Group %d with %d owners, %d admins and %d members\n",
		 SS_GROUP_ID(ss),
		 SS_GROUP_OWNER_CNT(ss),
		 SS_GROUP_ADMIN_CNT(ss),
		 SS_GROUP_SIZE(ss) );
  fprintf(stderr,"  Owners:");
  for ( i=0,max=SS_GROUP_OWNER_CNT(ss); i<max; ++i)
    fprintf(stderr," %s",ipHostStr(SS_GROUP_OWNER_ADDR(ss,i)));
  fprintf(stderr,"\n");
  /*
  fprintf(stderr,"  Admins:");
  for ( i=0,max=SS_GROUP_ADMIN_CNT(ss); i<max; ++i)
    fprintf(stderr," %s",ipHostStr(SS_GROUP_ADMIN(ss,i)));
  fprintf(stderr,"\n");
  */
  fprintf(stderr,"  Members:");
  for ( i=0,max=SS_GROUP_SIZE(ss); i<max; ++i)
    fprintf(stderr," %s",ipHostStr(SS_GROUP_MEMBER(ss,i)));
  fprintf(stderr,"\n");
}

static SState*
grpmemCreateSState( XObj sessn )
{
  SState* sstate = X_NEW(SState);

  SS_GROUP_ID(sstate)       = -1;
  SS_GROUP_SIZE(sstate)     = 0;
  SS_GROUP_SESSION(sstate)  = sessn;
  host_list_init( SS_GROUP_OWNER(sstate) );
  host_list_init( SS_GROUP_ADMINS(sstate) );
  host_list_init( SS_GROUP_MEMBERS(sstate) );
  return sstate;
}

static void
getproc_sessn(self)
    XObj self;
{
  self->pop     = (void*) grpmemPop;
  self->state   = (char *)grpmemCreateSState( self );
  self->control = grpmemControlSessn;
  self->close   = grpmemClose;
}

static void
grpmem_update( ss, cnt, hosts )
  SState*  ss;
  int      cnt;
  LOCALHOSTTYPE* hosts;
{
  int i;
  LOCALHOSTTYPE* ahost = hosts;
  LOCALHOSTTYPE* limit = hosts + cnt;

  xTrace1(grpmemp,TR_FULL_TRACE,"grpmem_update: %d addrs",cnt);

  SS_NEXT_QUERY(ss)     = QUERY_PERIOD+1;
  SS_REMAINING_QUERIES(ss) = MAX_QUERIES+1;

  host_list_clear( SS_GROUP_OWNER(ss) );

  for (i=0; (ahost<limit) && addr_is_valid(ahost); ++ahost) {
    add_group_owner( ss, ahost );
  }
  ++ahost;

  /*
  host_list_clear( SS_GROUP_ADMINS(ss) );

  for (i=0; (ahost<limit) && addr_is_valid(ahost); ++ahost)
    add_group_admin( ss, ahost );
  ++ahost;
  */

  host_list_clear( SS_GROUP_MEMBERS(ss) );

  for (i= cnt-(ahost-hosts); i>0; --i, ++ahost)
    add_group_member( ss, ahost );

  km_update( ss );

  xIfTrace(grpmemp,TR_FULL_TRACE)
    grpmemDumpSessn( ss );
}

static void
grpmem_addmem( ss, hosts )
  SState*  ss;
  LOCALHOSTTYPE* hosts;
{
  int i;
  int      cnt=0;
  LOCALHOSTTYPE* ahost;

  for (ahost=hosts;addr_is_valid(ahost);++ahost)
    ++cnt;

  xTrace1(grpmemp,TR_FULL_TRACE,"grpmem_addmem: %d new members",cnt);
  for ( i=0; i<cnt; ++i ) {
    if ( ( ahost = find_group_member( ss, &(hosts[i]) ) ) == NULL ) {
      xTrace1(grpmemp,TR_MAJOR_EVENTS,"grpmem_addmem: inserting %s",
		      ipHostStr(&(hosts[i])));
      add_group_member( ss, &(hosts[i]) );
    }
    else {
      xTrace1(grpmemp,TR_FULL_TRACE,"grpmem_addmem: %s already in group",
		      ipHostStr(&(hosts[i])));
    }
  }
}

static void
grpmem_delmem( ss, hosts )
  SState* ss;
  LOCALHOSTTYPE* hosts;
{
  LOCALHOSTTYPE* ahost;
  Msg*      msg;
  host_list del_members;

  xTrace0(grpmemp,TR_FULL_TRACE,"local_delmem");

  host_list_init( &del_members );

  for (; addr_is_valid(hosts); ++hosts) {

    /* used to create a list of removed members for notification */
    host_list_add( &del_members, hosts );

    if ( ( ahost = find_group_member( ss, hosts) ) != NULL ) {
      xTrace1(grpmemp,TR_MAJOR_EVENTS, "grpmem_delmem: found host %s, deleting",
		     ipHostStr(hosts));
      del_group_member( ss, ahost );
    }
  }

  /* Attempt to notify members removed from group */
  msg = build_delgrp_msg( ss );
  host_list_foreach( &del_members, (foreach_func_t*)grpmem_send_msg,
			    (void*)msg, (void*) ss );
  msgDestroy( msg );
  host_list_destroy( &del_members );

  xTrace0(grpmemp,TR_FULL_TRACE,"grpmem_delmem done.");
}

/*
 * the following functions are used by foreach_group_member:
 */

static int
foreach_cntmem( LOCALHOSTTYPE* ahost, int* cntmem )
{
  *cntmem += 1;
  return 0;
}

static int
foreach_allmem( LOCALHOSTTYPE* ahost, LOCALHOSTTYPE** buf )
{
  bcopy((char*)ahost,(char*)*buf,sizeof(LOCALHOSTTYPE));
  *buf = (*buf)+1;
  return 0;
}

static int
grpmem_cntmem( ss )
  SState* ss;
{
  int            hostcnt = 0;

  foreach_group_member( ss, (foreach_func_t*)foreach_cntmem, &hostcnt, NULL );

  return hostcnt;
}

static int
grpmem_allmem( ss, buf, len )
  SState* ss;
  LOCALHOSTTYPE* buf;
  int            len;
{
  LOCALHOSTTYPE* tmp    = buf;
  int            cntmem = 0;

  foreach_group_member( ss, (foreach_func_t*)foreach_cntmem, &cntmem, NULL );

  if ( (len/sizeof(LOCALHOSTTYPE)) < cntmem ) {
    xError("Buffer is too small to hold all hosts in group");
    return -1;
  }

  foreach_group_member( ss, (foreach_func_t*)foreach_allmem, &tmp, NULL );

  return cntmem;
}

static void
send_query( SState* ss )
{
  Msg* msg;
  xTrace0(grpmemp,TR_FULL_TRACE,"send_query: sending msg");
  msg = build_request_msg( ss );
  grpmem_send_msg( SS_GROUP_OWNER_ADDR(ss,0), msg, ss );
  msgDestroy( msg );
}

static void
send_update( SState* ss )
{
  Msg*    msg;
  xTrace0(grpmemp,TR_FULL_TRACE,"send_update: sending msgs");
  msg = build_update_msg( ss );

# if USE_BRDCST
    /*
    foreach_group_owner( ss, (foreach_func_t*)grpmem_bcast_msg,
			     (void*)msg, (void*) ss );
    */
    grpmem_bcast_msg(&((PState*)SS_GROUP_PROTL(ss)->state)->grp_addr, msg, ss );
# else
    foreach_group_member( ss, (foreach_func_t*)grpmem_send_msg,
			      (void*)msg, (void*) ss );
# endif
  msgDestroy( msg );
}

static int
periodic_group_check( IDTYPE* gidptr, XObj sessn )
{
  SState* ss = (SState*) sessn->state;
  PState* ps = (PState*) SS_GROUP_PROTL(ss)->state;

  int i,is_owner=0;
  int interface_cnt = PS_IF_CNT(ps);

  xTrace1(grpmemp,TR_FULL_TRACE,"periodic_group_check %d", *gidptr );
  for (i=0;i<interface_cnt;++i) {
    if ( find_group_owner( ss, PS_INTERFACE(ps,i) ) ) {
      is_owner = 1;
      break;
    }
  }
  if ( is_owner ) {
    send_update( ss );
  }
  else {
    --SS_NEXT_QUERY(ss);
    if ( SS_NEXT_QUERY(ss) == 0 ) {
#     if ( MAX_QUERIES )
	--SS_REMAINING_QUERIES(ss);
	if ( !SS_REMAINING_QUERIES(ss) ) {
	  xTrace1(grpmemp,TR_MAJOR_EVENTS,"no updates received for group %d, "
					 "deleting it", SS_GROUP_ID(ss) );
	  local_delgrp( sessn );
	}
	else
#     endif
	{
	  SS_NEXT_QUERY(ss) = QUERY_PERIOD;
	  send_query( ss );
	}
    }
  }

  return MFE_CONTINUE;
}

static void
km_update( SState* ss )
{
  XObj kms;
  Part p;
  XObj myprotl = SS_GROUP_PROTL(ss);
  LOCALHOSTTYPE kmcontents[GRPMEM_MAX_PARTS];

  partInit(&p,1);
  partPush( p, &SS_GROUP_ID(ss), sizeof(IDTYPE) );
  kms = xOpen( myprotl, myprotl, xGetDown(myprotl,KM_PROTL), &p );
  if ( kms == ERR_XOBJ ) {
    xError("grpmem: unable to open key manager session");
  }
  xControl(kms,KM_SET,(void*)(SS_GROUP_MEMBERS(ss)->hosts),
		      GRPMEM_MAX_PARTS * sizeof(LOCALHOSTTYPE) );
  xControl(kms,KM_RESOLVE, (void*) kmcontents, sizeof(kmcontents) );

  if ( bcmp( (char*)SS_GROUP_MEMBERS(ss)->hosts, (char*)kmcontents, GRPMEM_MAX_PARTS * sizeof(LOCALHOSTTYPE) ) )
    fprintf(stderr, "KM error!\n");

  xClose(kms);
}

static void
km_invalidate( SState* ss )
{
  XObj kms;
  Part p;
  XObj myprotl = SS_GROUP_PROTL(ss);

  partInit(&p,1);
  partPush( p, &SS_GROUP_ID(ss), sizeof(IDTYPE) );
  kms = xOpen( myprotl, myprotl, xGetDown(myprotl,KM_PROTL), &p );
  if ( kms == ERR_XOBJ ) {
    xError("grpmem: unable to open key manager session");
  }
  xControl(kms,KM_INVALIDATE,NULL,0);
  xClose(kms);
}

/*
 * update_event
 */

static void
update_event( Event e, XObj protl )
{
  PState* ps = (PState*) protl->state;
  mapForEach( ps->group_map, (void*)periodic_group_check, NULL );
  evSchedule( (EvFunc)update_event, (void*)protl, ps->update_interval );
}


/*
 *  grpmemControlSessn
 *
 */
static int
grpmemControlSessn( self, opcode, buf, len )
  XObj self;
  int opcode, len;
  char *buf;
{
  int  retval = 0;
  SState* ss = (SState*)self->state;

  xAssert(xIsSession(self));

  xTrace1(grpmemp, TR_MAJOR_EVENTS, "grpmemControlSessn, op = %d", opcode);

  switch(opcode) {

    case GRPMEM_ADDMEM: { /* buf: an array of LOCALHOSTYPE with host addrs */

      xTrace0(grpmemp,TR_FULL_TRACE, "grpmemControlSessn: GRPMEM_ADDMEM");

      grpmem_addmem( ss, (LOCALHOSTTYPE*) buf );

      km_update( ss );

      send_update( ss );

      return retval;
    }

    case GRPMEM_DELMEM: { /* buf: an array of LOCALHOSTYPE with host addrs */

      xTrace0(grpmemp,TR_FULL_TRACE, "grpmemControlSessn: GRPMEM_DELMEM");

      grpmem_delmem( ss, (LOCALHOSTTYPE*) buf );

      km_update( ss );

      send_update( ss );

      return retval;
    }

    case GRPMEM_CNTMEM: { /* buf: pointer to integer */
      checkLen(len,sizeof(int));

      xTrace0(grpmemp,TR_FULL_TRACE, "grpmemControlSessn: GRPMEM_CNTMEM");

      retval = grpmem_cntmem( ss );
      if ( buf )
	*((int*)buf) = retval;
      return retval;
    }

    case GRPMEM_ALLMEM: { /* buf: an empty array of LOCALHOSTYPE */

      xTrace0(grpmemp,TR_FULL_TRACE, "grpmemControlSessn: GRPMEM_ALLMEM");

      retval = grpmem_allmem( ss, (LOCALHOSTTYPE*)buf, len );
      return retval;
    }

#if 0
    case GRPMEM_DELMGR: { /* buf: an empty array of LOCALHOSTYPE */

      xTrace0(grpmemp,TR_FULL_TRACE, "grpmemControlSessn:  GRPMEM_DELMGR");

      xError("GRPMEM_DELMGR not implemented");
      return 0;
    }

    case GRPMEM_CNTMGR: { /* buf: pointer to integer */

      xTrace0(grpmemp,TR_FULL_TRACE, "grpmemControlSessn:  GRPMEM_CNTMGR");

      xError("GRPMEM_CNTMGR not implemented");
      return 0;
    }

    case GRPMEM_ALLMGR: { /* buf: an empty array of LOCALHOSTYPE */

      xTrace0(grpmemp,TR_FULL_TRACE, "grpmemControlSessn:  GRPMEM_ALLMGR");

      xError("GRPMEM_ALLMGR not implemented");
      return 0;
    }
#endif

    case GRPMEM_CNTOWN: { /* buf: pointer to integer */

      xTrace0(grpmemp,TR_FULL_TRACE, "grpmemControlSessn:  GRPMEM_CNTOWN");

      xError("GRPMEM_CNTOWN not implemented");
      return 0;
    }

    case GRPMEM_GETOWN: { /* buf: an empty array of LOCALHOSTYPE */

      xTrace0(grpmemp,TR_FULL_TRACE, "grpmemControlSessn:  GRPMEM_GETOWN");

    }

    default:
      return xControl(xGetDown(self, TRANSPORT_PROTL), opcode, buf, len);
  }
}

static void
local_delgrp( sessn )
  XObj sessn;
{
  SState* ss = (SState*)sessn->state;
  PState* ps    = (PState*)SS_GROUP_PROTL(ss)->state;
  IDTYPE  grpid = SS_GROUP_ID(ss);
  xTrace1(grpmemp,TR_MAJOR_EVENTS,"local_delgrp: deleting group %d",grpid);
  sessn->rcnt = 0;

  xTrace1(grpmemp,TR_MAJOR_EVENTS,"local_delgrp: group %d",grpid);
  mapUnbind( ps->group_map, (void*)&grpid );
  host_list_destroy( SS_GROUP_OWNER(ss) );
  host_list_destroy( SS_GROUP_ADMINS(ss) );
  host_list_destroy( SS_GROUP_MEMBERS(ss) );
  km_invalidate(ss);
  xClose(sessn);
}

/*
 *  grpmemControlProtl
 *
 */

/*
 * The following functions are used when listing a mapping:
 */

static int
map_foreach_cntgrp( IDTYPE *grpid, XObj grpsessn, int* cntgrp )
{
  xTrace1(grpmemp,TR_FULL_TRACE,"map_foreach_cntgrp called with cntgrp = %d",*cntgrp);
  *cntgrp += 1;
  return MFE_CONTINUE;
}

static int
map_foreach_allgrp( IDTYPE* grpid, XObj grpsessn, IDTYPE** allgrp )
{
  xTrace2(grpmemp,TR_FULL_TRACE,"map_foreach_allgrp called with grpid = %d, old allgrp is %d",*grpid,**allgrp);
  **allgrp = *grpid;
  *allgrp = (*allgrp)+1;
  return MFE_CONTINUE;
}

static void
local_allgrp( self, buf )
  XObj   self;
  IDTYPE *buf;
{
  IDTYPE* tmp = buf;
  PState* ps = (PState*)self->state;
  mapForEach( ps->group_map, (void*)map_foreach_allgrp, (void*)&tmp );
}

static int
local_cntgrp( self )
  XObj self;
{
  PState* ps = (PState*)self->state;
  int cntgrp = 0;
  mapForEach( ps->group_map, (void*)map_foreach_cntgrp, (void*)&cntgrp );
  return cntgrp;
}

static int
grpmemControlProtl( self, opcode, buf, len )
  XObj self;
  int opcode, len;
  char *buf;
{
  xTrace1(grpmemp, TR_MAJOR_EVENTS, "grpmemControlProtl, op = %d", opcode);

  xAssert(xIsProtocol(self));

  switch(opcode) {

    case GRPMEM_ALLGRP: {
      int cntgrp;

      xTrace0(grpmemp,TR_FULL_TRACE, "grpmemControlProtl: GRPMEM_ALLGRP");

      cntgrp = local_cntgrp( self );

      xTrace1(grpmemp,TR_FULL_TRACE,
	      "grpmemControlProtl: GRPMEM_ALLGRP found %d groups", cntgrp);

      if ( (len/sizeof(IDTYPE)) < cntgrp ) {
	xError("Buffer is too small for list of groups");
	return -1;
      }
      *(IDTYPE*)buf = 0;
      local_allgrp( self, (IDTYPE*)buf );
      return cntgrp;
    }

    case GRPMEM_CNTGRP: {
      int cntgrp;

      xTrace0(grpmemp,TR_FULL_TRACE, "grpmemControlProtl: GRPMEM_CNTGRP");

      checkLen(len,sizeof(int));
      cntgrp = local_cntgrp( self );
      if ( buf )
	*((int*)buf) = cntgrp;

      xTrace1(grpmemp,TR_FULL_TRACE,
	     "grpmemControlProtl: GRPMEM_CNTGRP found %d groups", cntgrp);

      return cntgrp;
    }

    case GRPMEM_DELGRP: {
      XObj    sessn;
      PState* ps = (PState*) self->state;
      SState* ss;
      IDTYPE  gid = *((IDTYPE*)(buf));

      xTrace1(grpmemp,TR_FULL_TRACE,
	      "grpmemControlProtl: in GRPMEM_DELGRP, gid = %d",gid);

      if (mapResolve(ps->group_map,(void*)&gid,(void**)&sessn)
	 ==XK_FAILURE) {
	xTrace1(grpmemp,TR_ERRORS,"grpmemControlProtl: no group %d",gid);
	return XK_FAILURE;
      }
      ss = (SState*) sessn->state;

      /*
      verify_local_ownership( ss, "GRPMEM_DELGRP" );
      */
      fprintf(stderr,"verify_local_ownership bypassed in GRPMEM_DELGRP\n");

      global_delgrp( self, sessn );

      return XK_SUCCESS;
    }

    default:
      return xControl(xGetDown(self, TRANSPORT_PROTL), opcode, buf, len);
  }
}

/* translate a command string to a command index */

static grpmem_cmd_codes
cmd_to_int(str)
  char* str;
{
  if (!strncmp(str, "grpadd", 8))
    return GRP_ADD;
  if (!strncmp(str, "grpdel", 8))
    return GRP_DEL;
  if (!strncmp(str, "grpowner", 8))
    return GRP_OWNER;
  if (!strncmp(str, "grpaddmem", 11))
    return GRP_ADD_MEM;
  if (!strncmp(str, "grpdelmem", 11))
    return GRP_DEL_MEM;
  if (!strncmp(str, "grprequest",12))
    return GRP_REQUEST;
  if (!strncmp(str, "grpreply",10))
    return GRP_REPLY;
  if (!strncmp(str, "grpaddadm",11))
    return GRP_ADD_ADM;
  if (!strncmp(str, "grpdeladm",11))
    return GRP_DEL_ADM;
  return -1;
}

static XObjRomOpt rmopts[] = {
			       { "rule", 2, romrule },
			       { "port", 1, romport },
			       { "update", 1, romupdate },
			       { "address", 1, romaddress },
			     };

/*
 *  romport
 *
 *   expect the definition of the port to be used:
 *
 *   grpmem  port port_number
 *
 */

static xkern_return_t
romport(self, ptr, fields, line, arg)
  XObj self;
  char **ptr;
  int  fields, line;
  void *arg;
{
  PState* ps = (PState*) self->state;
  xTrace0(grpmemp, TR_MAJOR_EVENTS, "grpmem PORT rom opt");

  return sscanf(ptr[2],"%d",&PS_CNTRL_PORT(ps))<1? XK_FAILURE: XK_SUCCESS;
}

static xkern_return_t
romupdate(self, ptr, fields, line, arg)
  XObj self;
  char **ptr;
  int  fields, line;
  void *arg;
{
  int retval;
  PState* ps = (PState*) self->state;
  xTrace0(grpmemp, TR_MAJOR_EVENTS, "grpmem UPDATE rom opt");

  retval = sscanf(ptr[2],"%d",&(ps->update_interval));
  ps->update_interval *= 1000000;

  return retval<1? XK_FAILURE: XK_SUCCESS;
}

/*
 *  romrule
 *
 *   expect to add a group like this
 *
 *   grpmem  rule grpadd grp_id [ list of host addrs ]
 *   grpmem  rule grpadm grp_id [ list of admin addrs ] (not yet)
 *   grpmem  rule grpmem grp_id [ list of member addrs ]
 *
 */

/* some convenient defines for working with rom entries */

#define ROM_CMD   2
#define ROM_GID   3
#define ROM_ADDR  4

static xkern_return_t
romrule(self, fields, numfields, line, arg)
  XObj self;
  char **fields;
  int  numfields, line;
  void *arg;
{
  PState *ps = (PState *)(self->state);
  SState *ss;
  XObj    sessn = NULL;
  LOCALHOSTTYPE work_area[GRPMEM_MAX_PARTS+1];
  LOCALHOSTTYPE *hosts=work_area;
  int     i,host_cnt;
  IDTYPE  rom_gid;
  grpmem_cmd_codes command;


  rom_gid = atoi(fields[ROM_GID]);
  command = cmd_to_int(fields[ROM_CMD]);

  xTrace3(grpmemp, TR_MAJOR_EVENTS, "grpmem RULE rom opt: %s(%d) %d",
	  fields[ROM_CMD], command , rom_gid );

  host_cnt = numfields - ROM_ADDR;
  if ( (host_cnt==1) && !strcmp(fields[ROM_ADDR],"owner") )
    host_cnt = 0;

  xTrace2(grpmemp,TR_NEVER,"rule in line %d has %d addrs",line, host_cnt);
  for ( i=0, fields+=ROM_ADDR ; i<host_cnt; ++i) {
    if ( str2ipHost( &(hosts[i]), fields[i] ) == XK_FAILURE ) {
      xTrace2(grpmemp, TR_ERRORS,
	      "invalid IP addr at line %d: %s", line, fields[i] );
    }
  }

  if ( command < 0 ) {
    xTrace1(grpmemp,TR_ERRORS,"grpmem rule: invalid command on line %d",line );
    return XK_FAILURE;
  }
  
  if (mapResolve(ps->group_map,(void*)&rom_gid,(void**)&sessn)
       == XK_FAILURE ) {
    sessn = grpmemCreateSessn(self,NULL,NULL,rom_gid,0,NULL);
    ss    = (SState*)sessn->state;
    km_update(ss);
  }
  ss = sessn->state;
  if ( host_cnt == 0 ) {
    host_cnt = PS_IF_CNT(ps); 
    hosts    = PS_INTERFACES(ps);
  }
  /*
  else if ( command == GRP_ADD ) {
     xTrace1(grpmemp, TR_ALWAYS,
     "GRP_ADD: fields after group id will be descarded on line %d", line );
  }
  */
  switch( command ) {
    /*
    case GRP_ADD_ADM:
      while(host_cnt--) {
	add_group_admin( ss, &(hosts[host_cnt]) );
      }
      break;
    */

    case GRP_ADD:
    /*  break; */

    case GRP_ADD_MEM:
      while(host_cnt--) {
	add_group_member( ss, &(hosts[host_cnt]) );
      }
      break;
    
    default:
      xTrace1(grpmemp, TR_ERRORS, "grpmem bogus opcode for rom opt %s",
	      fields[ROM_CMD]);
      break;
  }
  xIfTrace(grpmemp,TR_NEVER) {
    grpmemDumpSessn(ss);
  }

  return XK_SUCCESS;
}

static xkern_return_t
romaddress(self, ptr, fields, line, arg)
  XObj self;
  char **ptr;
  int  fields, line;
  void *arg;
{
  PState* ps = (PState*) self->state;

  xTrace1(grpmemp, TR_MAJOR_EVENTS, "grpmem ADDRESS rom opt at line %d",line);

  if ( str2ipHost( &(ps->grp_addr), ptr[2] ) == XK_FAILURE ) {
    xTrace2(grpmemp, TR_ERRORS, "invalid addr at line %d: %s", line, ptr[2] );
    return XK_FAILURE;
  }
  else {
    xTrace1(grpmemp, TR_MAJOR_EVENTS, "grpmem ADDRESS is %s",
	    ipHostStr(&(ps->grp_addr)));
  }
  return XK_SUCCESS;
}

void
grpmem_init( self )
  XObj self;
{
  PState	*ps;
  char	buffer[GRPMEM_MAX_PORTSIZE];
  Part	p[2];
  XObj  transport;

  xTrace0(grpmemp, TR_GROSS_EVENTS, "grpmem_init");
  xAssert(xIsProtocol(self));
  ps = X_NEW(PState);
  self->state = (VOID *)ps;
  ps->grpmem_portsize = xControl(xGetDown(self, KM_PROTL),
				  KM_HOSTSIZE, buffer, sizeof(buffer));
  if (ps->grpmem_portsize < 1) {
    xTrace1(grpmemp, TR_ERRORS, "grpmem init: Invalid key size %d", ps->grpmem_portsize);
    return ;
  }
  transport = xGetDown(self, TRANSPORT_PROTL);
  if ( ! xIsProtocol(transport) ) {
    xTrace0(grpmemp, TR_ERRORS, "grpmem_init: can't find transport protocol");
    xError("GRPMEM init errors -- not initializing");
    return;
  } else {
    xTrace1(grpmemp, TR_FULL_TRACE, "grpmem_init: transport is %s",
	    transport->name);
  }
  ps->group_map = mapCreate(GRPMEM_MAP_SZ,sizeof(IDTYPE));
  ps->passive   = mapCreate(GRPMEM_MAP_SZ,sizeof(Enable));
  ps->update_interval = UPDATE_INTERVAL;

  self->control     = grpmemControlProtl;
  self->open        = grpmemOpen;
  self->openenable  = grpmemOpenEnable;
  self->opendisable = grpmemOpenDisable;
  self->opendone    = grpmemOpenDone;
  self->demux       = grpmemDemux;

  xControl(transport,GETMYHOSTCOUNT,(void*)&PS_IF_CNT(ps),
		     sizeof(PS_IF_CNT(ps)));
  if ( PS_IF_CNT(ps) > MAX_HOST_COUNT ) {
    xError("Host has more interfaces than GRPMEM can handle");
    return;
  }
  xControl(transport,GETMYHOST,(void*)PS_INTERFACES(ps),sizeof(PS_INTERFACES(ps)));

  PS_CNTRL_PORT(ps) = -1;
  findXObjRomOpts(self, rmopts, sizeof(rmopts)/sizeof(XObjRomOpt), 0);
  if ( PS_CNTRL_PORT(ps) < 0 ) {
    xError("No valid port specified for grpoup membership protocol");
  }

  /* be ready to accept any message for this service */
  set_part_recv(p, ps);
  evSchedule( (EvFunc)update_event, (void*)self, 0 );
  xOpenEnable(self, self, transport, p);
}

static int
grpmem_bcast_msg( LOCALHOSTTYPE* interface, Msg* orig, SState* ss )
{
  Msg  msg;
  Part p[2];
  XObj lls, transport;
  XObj myprotl = SS_GROUP_PROTL(ss);
  LOCALHOSTTYPE* destn = interface;
  /*
  LOCALHOSTTYPE* destn = bcast_addr( interface );
  */
  /*
  destn = &(((PState*)myprotl->state)->grp_addr);
  */

  xTrace1(grpmemp, TR_MAJOR_EVENTS, "grpmem_bcast_msg to %s", ipHostStr(destn));
  msgConstructCopy( &msg, orig );
  set_part_send( p, destn, (PState*)myprotl->state );
  transport = xGetDown(myprotl, TRANSPORT_PROTL);
  lls = xOpen( myprotl, myprotl, transport, p  );
  xAssert(xIsSession(lls));
  xPush( lls, &msg );
  msgDestroy(&msg);

  return 0;
}

static int
grpmem_send_msg( LOCALHOSTTYPE* host, Msg* orig, SState* ss )
{
  Msg  msg;
  Part p[2];
  XObj lls, transport;
  XObj myprotl = SS_GROUP_PROTL(ss);

  xTrace1(grpmemp, TR_FULL_TRACE, "grpmem_send_msg to %s", ipHostStr(host));
  if ( its_myself( (PState*) myprotl->state, host ) ) {
    xTrace0(grpmemp, TR_FULL_TRACE, "grpmem_send_msg: local address, returning");
    return 0;
  }
  msgConstructCopy( &msg, orig );
  set_part_send( p, host, (PState*)myprotl->state );
  xTrace0(grpmemp, TR_FULL_TRACE, "grpmem_send_msg: openning transport session");
  transport = xGetDown(myprotl, TRANSPORT_PROTL);
  lls = xOpen( myprotl, myprotl, transport, p  );
  xAssert(xIsSession(lls));
  xTrace0(grpmemp, TR_FULL_TRACE, "grpmem_send_msg: pushing message");
  xPush( lls, &msg );
  msgDestroy(&msg);

  return 0;
}

static void
global_delgrp( XObj protl, XObj sessn )
{
  SState* ss         = (SState*) sessn->state;
  Msg*    delgrp_msg = build_delgrp_msg( ss );

  xTrace0(grpmemp,TR_FULL_TRACE, "in global_delgrp, sending messages");

  foreach_group_member( ss, (foreach_func_t*)grpmem_send_msg,
			    (void*) delgrp_msg, (void*) ss );

  msgDestroy( delgrp_msg );

  local_delgrp( sessn ); /* last operation */
}

