/*
 * $RCSfile: grpmem_i.h,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Log: grpmem_i.h,v $
 * Revision 1.1  1995/10/17 22:24:02  ho
 * Initial revision
 *
 * Revision 1.2  1995/04/28  00:21:08  dorgival
 * Minor syntax changes to make the IRIX compiler happy
 *
 * Revision 1.1  1995/04/24  18:48:15  dorgival
 * Initial revision
 *
 * Revision 1.1  1995/04/05  17:52:37  dorgival
 * Initial revision
 *
 *
 */

#define KM_PROTL  1
#define TRANSPORT_PROTL 0

#define REMOTE 0

/*
 * The following macros define the periodic behavior of the protocol:
 *
 * UPDATE_INTERVAL      Time (in microseconds) between periodic updates
 * QUERY_PERIOD         Number of update intervals between query messages
 *                      in case updates are not received from owner
 * MAX_QUERIES          Number of queries sent before group is considered
 *                      non-existent and is locally removed. A zero here
 *                      means "keep forever", do not remove.
 */

#define UPDATE_INTERVAL 10000000
#define MAX_QUERIES      2
#define QUERY_PERIOD     2

#define LOCALHOST_COMP_TYPE short
#define MAX_HOST_COUNT 4          /* Maximum number of interfaces in a host */

typedef enum { GRP_UPDATE,
	       GRP_ADD, GRP_DEL, GRP_OWNER,
               GRP_REQUEST, GRP_REPLY,
               GRP_ADD_MEM, GRP_DEL_MEM,
               GRP_ADD_ADM, GRP_DEL_ADM
	     } grpmem_cmd_codes;

#define MAX_CONTROL_BUF 100

extern tracegrpmemp;

typedef char * grpmem_rule;

typedef struct {
  int           grpmem_cntl_port;
  Map           passive;
  Map           group_map;       /* maps group names to sessions */
  int           grpmem_portsize; /* inherited from key manager   */
  int           update_interval;
  LOCALHOSTTYPE grp_addr;
  int           my_host_count;
  LOCALHOSTTYPE my_host[MAX_HOST_COUNT];
} PState;

#define PS_TRANSPORT(ps)   ((ps)->transport)
#define PS_KM(ps)          ((ps)->km)
#define PS_CNTRL_PORT(ps)  ((ps)->grpmem_cntl_port)
#define PS_IF_CNT(ps)      ((ps)->my_host_count)
#define PS_INTERFACES(ps)  ((ps)->my_host)
#define PS_INTERFACE(ps,idx) (&((PS_INTERFACES(ps))[idx]))


#define IDTYPE  long

typedef struct {
  IDTYPE    group_id;
  host_list group_owner;
  host_list group_members;
  host_list group_admins;
  XObj	    kms;
  XObj      session;
  int       next_query;
  int       remaining_queries;
} SState;

#define SS_GROUP_ID(ss)         ((ss)->group_id)
#define SS_GROUP_OWNER(ss)        (&((ss)->group_owner))
#define SS_GROUP_OWNER_ADDR(ss,idx) (HOST_LIST_HOST(SS_GROUP_OWNER(ss),idx))
#define SS_GROUP_OWNER_CNT(ss)    (HOST_LIST_SIZE(SS_GROUP_OWNER(ss)))
#define SS_GROUP_MEMBERS(ss)      (&((ss)->group_members))
#define SS_GROUP_MEMBER(ss,idx)   (HOST_LIST_HOST(SS_GROUP_MEMBERS(ss),idx))
#define SS_GROUP_SIZE(ss)         (HOST_LIST_SIZE(SS_GROUP_MEMBERS(ss)))
#define SS_GROUP_ADMINS(ss)       (&((ss)->group_admins))
#define SS_GROUP_ADMIN(ss,idx)    (HOST_LIST_HOST(SS_GROUP_ADMINS(ss),idx))
#define SS_GROUP_ADMIN_CNT(ss)    (HOST_LIST_SIZE(SS_GROUP_ADMINS(ss)))
#define SS_KMS(ss)                ((ss)->kms)
#define SS_BINDING(ss)            ((ss)->binding)
#define SS_GROUP_SESSION(ss)      ((ss)->session)
#define SS_GROUP_PROTL(ss)        (SS_GROUP_SESSION(ss)->myprotl)
#define SS_NEXT_QUERY(ss)         ((ss)->next_query)
#define SS_REMAINING_QUERIES(ss)  ((ss)->remaining_queries)

typedef struct {
  unsigned char cmd;
  IDTYPE   group_id;
  int      count;
  LOCALHOSTTYPE hosts[1];
} grpmem_msg;

#define MSG_COM(msg)        ((msg)->cmd)
#define MSG_GID(msg)        ((msg)->group_id)
#define MSG_HOST_COUNT(msg) ((msg)->count)
#define MSG_HOSTS(msg)      ((msg)->hosts)
#define MSG_HOST(msg,idx)   (&((msg)->hosts[idx]))
#define MSG_SIZE(hosts)     (sizeof(grpmem_msg)+((hosts)*sizeof(LOCALHOSTTYPE)))
#define MSG_NEW(hosts)      ((grpmem_msg*)xMalloc(MSG_SIZE(hosts)))

