/*
 * 
 * ipsec application interface protocol
 *     a simple intermediary for catching connection opens
 *     and attempting to turn them into secure connections
 *
 * $Log: spiint.c,v $
 * Revision 1.8  1997/03/10 22:11:56  mjk
 * switched to xTraceP?() macros.
 *
 *
 */

#include "xkernel.h"
#include "saidmgr.h"
#include "spiint_i.h"

int tracespiintp;

static XObj upperprot = ERR_XOBJ;

#define TRANSPORT_PROTL 0
#define SAIDMGR_PROTL   1
#define MAX_SPI_PAIRS  100

void spiint_init(XObj);
static int  spiintControl(XObj, int, char *, int);
static XObj spiintOpen( XObj, XObj, XObj, Part *);
static xkern_return_t spiintOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t spiintOpenEnable( XObj, XObj, XObj, Part *);

static xkern_return_t romusephoturis(XObj, char **, int, int, void *);
static xkern_return_t romalgorithm(XObj, char **, int, int, void *);

static XObjRomOpt rmopts[] = {{ "usephoturis", 5, romusephoturis },
			      { "algorithm", 4, romalgorithm }
				 };

static xkern_return_t
romalgorithm(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  PState * ps = (PState *) self->state;
  if (fields > 3)
  {
    ps->allowed_algo = xControl(xGetDown(self,SAIDMGR_PROTL),SAIDMGR_RESOLVE_NAME,ptr[2],strlen(ptr[2]));
    if(strcmp(ptr[3],"esp_tunnel")==0)
    	ps->allowed_mode = 1;
    else if(strcmp(ptr[3],"esp_encaps")==0)
    	ps->allowed_mode = 2;
    else if(strcmp(ptr[3],"ah")==0)
    	ps->allowed_mode = 0;
    else 
    	Kabort("Unknown IPSEC mode");
  }
  else
    Kabort("No allowed algorithm and mode set");
  return XK_SUCCESS;
}

static xkern_return_t
romusephoturis(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  PState * ps = (PState *) self->state;
  if (fields > 4)
  {
    ps->user_name = xMalloc(strlen(ptr[2])+1);
    memcpy(((char *)(ps->user_name)),ptr[2],strlen(ptr[2])+1);
    ps->neg_algo = xControl(xGetDown(self,SAIDMGR_PROTL),SAIDMGR_RESOLVE_NAME,ptr[3],strlen(ptr[3]));
    if(strcmp(ptr[4],"esp_tunnel")==0)
    	ps->neg_mode = 1;
    else if(strcmp(ptr[4],"esp_encaps")==0)
    	ps->neg_mode = 2;
    else if(strcmp(ptr[4],"ah")==0)
    	ps->neg_mode = 0;
    else 
    	Kabort("Unknown IPSEC mode");
  }
  else
    Kabort("No server identity, algorithm and mode set");
  xTraceP3(self, TR_MAJOR_EVENTS, "romoption usephoturis id: %s %u %u",
	   ps->user_name,ps->neg_algo,ps->neg_mode);

  return XK_SUCCESS;
}

void 
spiint_init(self)
XObj self;
{
  PState * ps;
  self->open = spiintOpen;
  self->opendone = spiintOpenDone;
  self->openenable = spiintOpenEnable;
  self->control = spiintControl;
  self->state = xMalloc(sizeof(PState));
  ps = (PState *) self->state;
  ps->user_name = NULL;
  ps->neg_algo = -1;
  ps->neg_mode = -1;
  ps->allowed_algo =-1;
  ps->allowed_mode=-1;
  findXObjRomOpts(self, rmopts, sizeof(rmopts)/sizeof(rmopts[0]), 0);
}



static int
spiintControl(s, op, b, l)
     XObj s;
     char *b;
     int op;
     int l;
{
  if (op == SAIDMGR_RESCAN_PRESET_KEY_FILE) {
   return 
     xControl(xGetDown(s,SAIDMGR_PROTL),SAIDMGR_RESCAN_PRESET_KEY_FILE,NULL,0);
  }
  return xControl(xGetDown(s, TRANSPORT_PROTL), op, b, l);
}

#define PART_TYPE_SECURE 4

static XObj
spiintOpen(self, hlp, hlptype, p)
     XObj self, hlp, hlptype;
     Part *p;
{
  XSaid buf[MAX_SPI_PAIRS*2];
  int k;
  int y;
  int ret;
  IPhost *peer1 = 0;
  IPhost *peer2 = 0;
  static int   ptysec = - PART_TYPE_SECURE;
  PState * ps = (PState *) self->myprotl->state;

  if (upperprot == ERR_XOBJ) upperprot = hlp;

  if (partLen(p) > 2) 
    {
      xTraceP0(self, TR_MAJOR_EVENTS,
	       "opening a session using SPI in part list");
      return xOpen(hlp, hlptype, xGetDown(self,TRANSPORT_PROTL), p);
    } 


  /* the remote machine address must be in the buf */

  peer1 = partPop(p[REMOTE_PART]);
  if (partStackTopByteLen(p[REMOTE_PART]) >= 0)
    peer2 = partPop(p[REMOTE_PART]);
  bcopy((char *)(peer2?peer2:peer1), (char *)buf, sizeof(IPhost));
  if (peer2) partPush(p[REMOTE_PART], peer2, sizeof(IPhost));
  if (peer1) partPush(p[REMOTE_PART], peer1, sizeof(IPhost));
  ret = xControl(xGetDown(self,SAIDMGR_PROTL),SAIDMGR_GET_SPIS_FOR_HOST,
		 (char*) buf, sizeof(buf));
  /* returns pair of SPIs in buf first the local then the remote spi.*/
  /* The ret value is the number of pairs returned. If there are */
  /* more SPI pairs as there is buffer size the buffer is filled */
  /* with arbitary SPI sets for the requested host. */
  if ((ret <= 0) && (ps->user_name == NULL))
    {
      xTraceP0(self, TR_MAJOR_EVENTS, "opening a non-secure session");
      return xOpen(hlp, hlptype, xGetDown(self,TRANSPORT_PROTL), p);
    } 
  else
    {
      int chosen = -1;
      if (ret > 0) 
      {
	   chosen = 0; /* INCOMPLETE: The SA should be selcted using the values of the 
			  algorithm romoption */
           y =0;
           for (k=0; k<ret ; k++)
		{
	  	   xTraceP2(self, TR_MAJOR_EVENTS, "local spi %u remote spi %u",
		  	buf[y],buf[y+1]);
	  	   y+=2;
	 	}
      }
      if (( chosen < 0 ) && (ps->user_name != NULL))
      {
		Part 	pp[3];
		XObj	lls;
		IPhost	clienthost;
	        IPhost  serverhost;
		serverhost = (peer2?*peer2:*peer1);
	  	xControl(xGetDown(self, TRANSPORT_PROTL), GETMYHOST, (char *)&clienthost, sizeof(IPhost));
                partInit(pp, 3);
    		partPush(pp[REMOTE_PART], &serverhost, sizeof(IPhost));
		partPush(pp[LOCAL_PART], &clienthost, sizeof(IPhost));
		if (strcmp (ps->user_name,"*")==0)
		{
			char 	b[20];
			char  * t;
			t = ipHostStr(&clienthost);
			strcpy(b,t);
			partPush(pp[SPI_PART], b, strlen(b)+1);
		}
		else
   			partPush(pp[SPI_PART], ps->user_name, strlen((char *) ps->user_name)+1); /* keep the trailing zero  */
   		partPush(pp[SPI_PART], &ps->neg_mode, sizeof(XIpsecMode));
    		partPush(pp[SPI_PART], &ps->neg_algo, sizeof(XAlgorithm));
 		lls = xOpen(self,self,xGetDown(self,SAIDMGR_PROTL),pp);
		if ( lls == ERR_XOBJ)
		{
	  	   	xTraceP0(self, TR_MAJOR_EVENTS,
				 "Key exchange failed open insecure connection");
      			return xOpen(hlp, hlptype, xGetDown(self,TRANSPORT_PROTL), p);
		}
		buf[0] = xControl(lls,SAIDMGR_GETLSAID,NULL,0);
		chosen=0;
      }
      if (chosen >= 0)
      {
	Part np[SPI_PART+1];
	XSaid spi = buf[chosen];
	IPhost myhost;
	
	partInit(np, SPI_PART+1);
	np[0] = p[0];
	np[0].len = SPI_PART+1;
	if (partLen(p)>1) {
	  np[1] = p[1];
	  np[1].len = SPI_PART+1;
	}
	else {
	  xControl(xGetDown(self, TRANSPORT_PROTL), GETMYHOST, (char *)&myhost, sizeof(IPhost));
	  partPush(np[LOCAL_PART], &myhost, sizeof(myhost));
	}
	partPush(np[SPI_PART], &spi, sizeof(XSaid *));
	partPush(np[SPI_PART], (void *)ptysec, 0);
	xTraceP2(self, TR_MAJOR_EVENTS, 
		"opening a secure session, spi %d 0x%x",
		spi, spi);
	return xOpen(hlp, hlptype, xGetDown(self,TRANSPORT_PROTL), np);
      }
      else
      {
 	 xTraceP0(self, TR_MAJOR_EVENTS,
		  "couldn't achive required security open insecure connection");
      	 return xOpen(hlp, hlptype, xGetDown(self,TRANSPORT_PROTL), p);
      }
    }
}

static xkern_return_t
spiintOpenDone( self, lls, llp, hlpType )
     XObj self, lls, llp, hlpType;
{
  char buf[100];
  Part p[SPI_PART+1];
  IPhost *peer;

  xControl(lls, GETPARTICIPANTS, buf, sizeof(buf));
  partInternalize(p, buf);
  peer = partPop(p[REMOTE_PART]);
  if (partLen(p)<=SPI_PART)
    xTraceP4(self, TR_ERRORS,
	   "receiving a non-secure connection request from host %d.%d.%d.%d",
	    peer->a, peer->b, peer->c, peer->d);
  xSetUp(lls, upperprot);
  return xOpenDone(upperprot, lls, llp);
}

static xkern_return_t
spiintOpenEnable(self, h, ht, p)
     XObj self, h, ht;
     Part *p;
{
  if (upperprot == ERR_XOBJ) upperprot = h;
  return xOpenEnable(h, ht, xGetDown(self, TRANSPORT_PROTL), p);
}
