/*
 * $RCSfile: tcp_internal.h,v $
 *
 * Derived from:
 *
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 *	@(#)tcp.h	7.3 (Berkeley) 12/7/87
 *
 * Modified for x-kernel v3.2
 * Modifications Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Log: tcp_internal.h,v $
 * Revision 1.1  1996/07/11 21:39:21  nicko
 * Initial revision
 *
 * Revision 1.40  1995/11/04  00:01:54  ho
 * Added code for dealing with multiple saids on the security participant.
 *
 * Revision 1.39  1995/06/07  21:43:31  spatsch
 * added ipsec support.
 *
 * Revision 1.38  1994/11/22  23:59:17  dorgival
 * Header structure was changed; endianess details now dealt with by usit.
 *
 * Revision 1.38  1994/11/18  22:03:04  dorgival
 * Header structure was changed; endianess details now dealt with by usit.
 *
 * Revision 1.37  1994/04/22  02:15:35  davidm
 * struct in_addr and tcp_seq are now of type xk_u_int32
 * instead of u_long.
 *
 * Revision 1.36  1993/12/16  01:41:38  menze
 * fixed '#else' comment
 *
 */

/**************************************************************************

tcp_internal.h

**************************************************************************/
#include "xkernel.h"
#include "sb.h"
#include "ip.h"
#include "tcp.h"
#include "tcp_port.h"

#ifndef ENDIAN
/*
 * Definitions for byte order,
 * according to byte significance from low address to high.
 */
#define	LITTLE	1234		/* least-significant byte first (vax) */
#define	BIG	4321		/* most-significant byte first */
#define	PDP	3412		/* LSB first in word, MSW first in long (pdp) */
#if defined(vax) || defined(pmax)
#define	ENDIAN	LITTLE
#else
#define	ENDIAN	BIG		/* byte order on mc68000, tahoe, most others */
#endif
#endif /* ENDIAN */

typedef u_long 		n_time;
#undef KPROF
#define PR_SLOWHZ	2
#define PR_FASTHZ	5
/* #define KERNEL */
#define BSD 43
#define TCP_S_COMPAT_42
#ifndef MIN
#define MIN(A,B) ((A) <= (B) ? (A) : (B))
#endif
#ifndef MAX
#define MAX(A,B) ((A) >= (B) ? (A) : (B))
#endif

typedef struct {
    Semaphore	s;
    int		waitCount;
} Tcp_SSem;

struct tcp_sstate {
    u_int	flags;		/* see below */
    struct inpcb *inpcb;
    struct tcp_scb *tcp_scb;
    Tcp_SSem	waiting;
    Tcp_SSem	lock;
    XObj	hlpType;	/* for passively opened sessions */
    struct sb	*snd;
    int		rcv_space;	/* amount of space in the receiver's buffer */
    int		rcv_hiwat;	/* size of receiver's buffer */
    int		closed;
};

#define TCP_SST_NBIO		0x0001	/* non-blocking i/o operations */
#define TCP_SST_OOBINLINE		0x0002	/* inline urgent data */
#define TCP_SST_PUSH_ALWAYS	0x0004	/* assume TCP_S-push after every xPush */
#define TCP_SST_RCV_ACK_ALWAYS	0x0008	/* assume infinite receive buffer */

#define sotoinpcb(S) (((struct tcp_sstate *)((S)->state))->inpcb)
#define sototcp_sst(S) ((struct tcp_sstate *)(S)->state)
#define tcp_scbtoso(tp) (tp->t_inpcb->inp_session)
extern int tracetcp_sp;

/*
 * The arguments to usrreq are:
 *	(*protosw[].pr_usrreq)(up, req, m, nam, opt);
 * where up is a (struct socket *), req is one of these requests,
 * m is a optional mbuf chain containing a message,
 * nam is an optional mbuf chain containing an address,
 * and opt is a pointer to a socketopt structure or nil.
 * The protocol is responsible for disposal of the mbuf chain m,
 * the caller is responsible for any space held by nam and opt.
 * A non-zero return from usrreq gives an
 * UNIX error number which should be passed to higher level software.
 */
#define	PRU_ATTACH		0	/* attach protocol to up */
#define	PRU_DETACH		1	/* detach protocol from up */
#define	PRU_BIND		2	/* bind socket to address */
#define	PRU_LISTEN		3	/* listen for connection */
#define	PRU_CONNECT		4	/* establish connection to peer */
#define	PRU_ACCEPT		5	/* accept connection from peer */
#define	PRU_DISCONNECT		6	/* disconnect from peer */
#define	PRU_SHUTDOWN		7	/* won't send any more data */
#define	PRU_RCVD		8	/* have taken data; more room now */
#define	PRU_SEND		9	/* send this data */
#define	PRU_ABORT		10	/* abort (fast DISCONNECT, DETATCH) */
#define	PRU_CONTROL		11	/* control operations on protocol */
#define	PRU_SENSE		12	/* return status into m */
#define	PRU_RCVOOB		13	/* retrieve out of band data */
#define	PRU_SENDOOB		14	/* send out of band data */
#define	PRU_SOCKADDR		15	/* fetch socket's address */
#define	PRU_PEERADDR		16	/* fetch peer's address */
#define	PRU_CONNECT2		17	/* connect two sockets */
/* begin for protocols internal use */
#define	PRU_FASTTIMO		18	/* 200ms timeout */
#define	PRU_SLOWTIMO		19	/* 500ms timeout */
#define	PRU_PROTORCV		20	/* receive from below */
#define	PRU_PROTOSEND		21	/* send to below */

#if	!defined(KERNEL) || MACH_VMTP
#define PRU_INVOKE		22	/* Invoke a message transaction */
#define PRU_RECVREQ		23	/* Receive a request */
#define PRU_REPLY		24	/* Send a reply */
#define PRU_FORWARD		25	/* Forward a request */
#define PRU_PROBE		26	/* Probe an entity */
#define PRU_GETREPLY		27 	/* Get next reply (multicast) */
#define	PRU_NREQ		27
#else	/* !defined(KERNEL) || MACH_VMTP */
#define PRU_NREQ		21	
#endif /* !defined(KERNEL) || MACH_VMTP */


extern char *prurequests[];

/*
 * Subset of the inpcb defined in /usr/include/netinet/in_pcb.h
 */
#ifdef XKMACHKERNEL
struct in_addr
	{
		xk_u_int32 s_addr;
	} ;
#endif /* XKMACHKERNEL */

struct inpcb {
	struct	inpcb *inp_next,*inp_prev;
	struct	inpcb *inp_head;
	caddr_t	inp_ppcb;		/* pointer to per-protocol pcb */
	struct in_addr inp_laddr, inp_raddr;
	u_short	inp_lport, inp_rport;
	XObj	inp_session;
};

#define RCV 0
#define SND 1

#ifndef XKMACHKERNEL
#include <sys/errno.h>
#else
#define	EINVAL		22		/* Invalid argument */
#define	EWOULDBLOCK	35		/* Operation would block */
#define	EOPNOTSUPP	45		/* Operation not supported on socket */
#define ECONNRESET	54
#define ENOBUFS		55
#define	EISCONN		56		/* Socket is already connected */
#define ETIMEDOUT	60
#define ECONNREFUSED	61
#endif /* XKMACHKERNEL */

typedef	xk_u_int32	tcp_s_seq;
/*
 * TCP_S header.
 * Per RFC 793, September, 1981.
 */
struct tcp_shdr {
	u_short	th_sport;		/* source port */
	u_short	th_dport;		/* destination port */
	u_int	th_seq;			/* sequence number */
	u_int	th_ack;			/* acknowledgement number */
	u_int	th_off:4,		/* data offset */
		th_x2:4;		/* (unused) */
	u_char	th_flags;
#define	TH_FIN	0x01
#define	TH_SYN	0x02
#define	TH_RST	0x04
#define	TH_PUSH	0x08
#define	TH_ACK	0x10
#define	TH_URG	0x20
	u_short	th_win;			/* window */
	u_short	th_sum;			/* checksum */
	u_short	th_urp;			/* urgent pointer */
};

#define	TCP_SOPT_EOL	0
#define	TCP_SOPT_NOP	1
#define	TCP_SOPT_MAXSEG	2

/*
 * Default maximum segment size for TCP_S.
 * With an IP MSS of 576, this is 536,
 * but 512 is probably more convenient.
 */
#ifdef	lint
#define	TCP_S_MSS	536
#else
#ifndef IP_MSS
#define	IP_MSS	576
#endif
#define	TCP_S_MSS	MIN(512, IP_MSS - sizeof (struct tcp_siphdr))
#endif

/*
 * User-settable options (used with setsockopt).
 */
#define	TCP_S_NODELAY	0x01	/* don't delay send to coalesce packets */
#define	TCP_S_MAXSEG	0x02	/* set maximum segment size */


/*
 * X-Kernel defines
 */

/*====================  #8 10 lines added 070390 cliff ===================*/

#define TCP_S_IPSEC		/* Compiles with code necessary for IPSEC */
#define IPSEC		        /* Compiles with code necessary for RFC1825 */

typedef struct pstate  {
    IPhost	myHost;
    Map     	activeMap;
    Map      	passiveMap;
    VOID	*portstate;
    struct tcp_sstat	*tcp_sstat;	/* tcp_s statistics */
    struct inpcb	*tcb;		/* head of queue of active tcp_scb's */
#ifdef TCP_S_IPSEC
    int 	useIpsec;
#endif	
}  PSTATE;


typedef TCP_Sport  PassiveId;

typedef struct {
      unsigned short localport;
      unsigned short remoteport;
      IPhost         remoteaddr;
} ActiveId;

typedef struct {
	IPpseudoHdr	*h;
	Msg		*m;
} hdrStore_t;

#define FROM_ENABLE	-1
#define TCP_SRCVWIN	(TCP_S_BUFFER_SPACE)

extern char *tcp_sstates[];

#if defined(__STDC__) || defined(__GNUC__)

/*
 * in_hacks.c
 */
int	in_pcballoc( XObj );
void	in_pcbdisconnect( struct inpcb * );
void 	in_pcbdetach( struct inpcb * );
bool	in_broadcast( struct in_addr );
u_short	in_cksum( Msg *, u_short *, int len );

/*
 * tcp_s_subr.c
 */
struct tcp_siphdr *	tcp_s_template( struct tcp_scb * );
struct tcp_scb *	tcp_s_newtcp_scb( struct inpcb * );
struct tcp_scb *	tcp_s_drop( struct tcp_scb *, int errnum );
struct tcp_scb * 	tcp_s_destroy( struct tcp_scb * );
void	tcp_s_notify( struct inpcb * );
void	tcp_s_quench( struct inpcb * );
void	tcp_s_respond( struct tcp_scb *, struct tcp_shdr *, IPpseudoHdr *,
			     tcp_s_seq ack, tcp_s_seq seq, int flags,
		             XObj tcp_sProtl );
 
/*
 * tcp_s_hdr.c
 */
void 	tcp_sHdrStore(void *, char *, long, void *);
void 	tcp_sOptionsStore(void *, char *, long, void *);
long	tcp_sOptionsLoad(void *, char *, long, void *);
long 	tcp_sHdrLoad(void *, char *, long, void *);

/*
 * tcp_s_timer.c
 */
void	tcp_s_fasttimo( Event, void * );
void	tcp_s_slowtimo( Event, void * );
void	tcp_s_canceltimers( struct tcp_scb * );
struct tcp_scb *	tcp_s_timers( struct tcp_scb *, int );

/*
 * tcp_s_x.c
 */
void	delete_tcp_sstate( struct tcp_sstate * );
struct tcp_sstate	*	new_tcp_sstate( void );
void	sorwakeup( XObj );
void	sowwakeup( XObj );
void	soabort( XObj );
void	socantsendmore( XObj );
void	socantrcvmore( XObj );
void	sohasoutofband( XObj, u_int );
void	soisdisconnected( XObj );
void	soisdisconnecting( XObj );
void	soisconnected( XObj );
void	soisconnecting( XObj );
int	soreserve( XObj, int, int );
#ifdef TCP_S_IPSEC
XObj	sonewconn( XObj, XObj, XObj, XObj, IPhost *, IPhost *, int, int );
#else
XObj	sonewconn( XObj, XObj, XObj, IPhost *, IPhost *, int, int );
#endif
void	tcp_sSemWait( Tcp_SSem * );
void	tcp_sSemSignal( Tcp_SSem * );
void	tcp_sSemInit( Tcp_SSem *, int );
void	tcp_sVAll( Tcp_SSem * );

/* 
 * tcp_s_output.c
 */
int	tcp_s_output( struct tcp_scb * );
void	tcp_s_setpersist( struct tcp_scb * );

/* 
 * tcp_s_usrreq.c
 */
int	tcp_s_attach( XObj );
int	tcp_s_usrreq( XObj, int, Msg *, Msg * );
struct tcp_scb *tcp_s_disconnect(struct tcp_scb * );
struct tcp_scb *tcp_s_usrclosed( struct tcp_scb * );

/* 
 * tcp_s_input.c
 */
void	print_reass( struct tcp_scb *, char * );
xkern_return_t	tcp_s_input( XObj, XObj, Msg * );
int	tcp_s_mss( struct tcp_scb * );
xkern_return_t	tcp_sPop( XObj, XObj, Msg *, VOID * );
int	tcp_s_reass( struct tcp_scb *, struct tcp_shdr *, XObj, Msg *, Msg * );

/* 
 * tcp_s_debug.c
 */
void	tcp_s_trace(int, int, struct tcp_scb *, struct tcp_siphdr *, int );
char *	tcp_sFlagStr( int );

#else

/*
 * in_hacks.c
 */
int	in_pcballoc();
void	in_pcbdisconnect();
void 	in_pcbdetach();
bool	in_broadcast();
u_short	in_cksum();

/*
 * tcp_s_subr.c
 */
struct tcp_siphdr *	tcp_s_template();
struct tcp_scb *	tcp_s_newtcp_scb();
struct tcp_scb *	tcp_s_drop();
struct tcp_scb * 	tcp_s_destroy();
void	tcp_s_notify();
void	tcp_s_quench();
void	tcp_s_respond();
 
/*
 * tcp_s_hdr.c
 */
void 	tcp_sHdrStore();
void 	tcp_sOptionsStore();
long	tcp_sOptionsLoad();
long 	tcp_sHdrLoad();

/*
 * tcp_s_timer.c
 */
void	tcp_s_fasttimo();
void	tcp_s_slowtimo();
void	tcp_s_canceltimers();
struct tcp_scb *	tcp_s_timers();

/*
 * tcp_s_x.c
 */
void	delete_tcp_sstate();
struct tcp_sstate	*	new_tcp_sstate();
void	sorwakeup();
void	sowwakeup();
void	soabort();
void	socantsendmore();
void	socantrcvmore();
void	sohasoutofband();
void	soisdisconnected();
void	soisdisconnecting();
void	soisconnected();
void	soisconnecting();
int	soreserve();
XObj	sonewconn();
void	tcp_sSemWait();
void	tcp_sSemSignal();
void	tcp_sSemInit();
void	tcp_sVAll();

/* 
 * tcp_s_output.c
 */
int	tcp_s_output();
void	tcp_s_setpersist();

/* 
 * tcp_s_usrreq.c
 */
int	tcp_s_attach();
int	tcp_s_usrreq();
struct tcp_scb *tcp_s_disconnect();
struct tcp_scb *tcp_s_usrclosed();

/* 
 * tcp_s_input.c
 */
void	print_reass();
xkern_return_t	tcp_s_input();
int	tcp_s_mss();
xkern_return_t	tcp_sPop();
int	tcp_s_reass();

/* 
 * tcp_s_debug.c
 */
void	tcp_s_trace();
char *	tcp_sFlagStr();

#endif /* __STDC__ */

#include "insque.h"
