connect(X,Y) :- link(X,Y); Z^(link(X,Z), connect(Z,Y)).

all_hlp(X,S) :- setof(H,connect(H,X),S).
all_llp(X,S) :- setof(L,connect(X,L),S).
all_protocols(S) :- setof(P,protocol(P),S).

tlp(P) :- protocol(P), link(_,P), !, fail.
tlp(P) :- protocol(P).
blp(P) :- protocol(P), link(P,_), !, fail.
blp(P) :- protocol(P).

all_tlp(S) :- setof(P,(protocol(P), if(link(_,P), fail, true)),S).
all_blp(S) :- setof(P,(protocol(P), if(link(P,_), fail, true)),S).

security(P) :- attribute(P, privacy).
security(P) :- blp(P), !, fail.
security(P) :- link(P,X), if(security(X), fail, (!, fail)).
security(P) :- protocol(P).

syssec :- protocol(P), if(link(_,P), fail, if(security(P), fail, (!, fail))).
syssec.

integrity(P) :- attribute(P, integrity).
integrity(P) :- blp(P), !, fail.
integrity(P) :- link(P,X), if(integrity(X), fail, (!, fail)).
integrity(P) :- protocol(P).

sysint :- protocol(P), if(link(_,P), fail, if(integrity(P), fail, (!, fail))).
sysint.

integrity_or_security(P) :- integrity(P); security(P).
integrity_or_security(P) :- link(P,X), if(security_or_integrity(X), fail, (!, fail)).
integrity_or_security(P) :- protocol(P).

sysintsec :- protocol(P), if(link(_,P), fail, if(integrity_or_security(P), fail, (!, fail))).
sysintsec.

above_check :- above(X,Y), if(connect(Y,X), fail, (!, fail)).
above_check.

below_check :- below(X,Y), if(connect(X,Y), fail, (!, fail)).
below_check.

syscor :- above_check, below_check.

leafmode(P) :- blp(P), drivers(D), if(member(P,D), (!, fail), true).
leafmode(P) :- romopt([P,leafmode]).
leafmode(P) :- link(P,A), link(P,B), !, fail.
leafmode(P) :- link(P,KM), kms(L), member(KM,L).

leafpath(P) :- leafmode(P).
leafpath(P) :- link(P,X), if(leafpath(X), fail, (!, fail)).
leafpath(P) :- protocol(P), drivers(D), if(member(P,D), (!,fail), true).

member(X,[X|T]).
member(X,[_|T]) :- member(X,T).

drivers([simeth,ethdrv,ns8390,xklance,irix,irixeth,irixfddi,simfddi,lance]).
